"""XC density kernels for response function calculations."""

import numpy as np

from gpaw.response.localft import LocalFTCalculator
from gpaw.response.fxc_kernels import AdiabaticFXCCalculator


def get_density_xc_kernel(qpd, gs, context, functional='ALDA',
                          rshelmax=-1, rshewmin=None,
                          chi0_wGG=None):
    """Density-density xc kernels.
    Factory function that calls the relevant functions below."""

    p = context.print
    nspins = len(gs.nt_sR)
    assert nspins == 1

    if functional[0] == 'A':
        # Standard adiabatic kernel
        p('Calculating %s kernel' % functional)
        localft_calc = LocalFTCalculator.from_rshe_parameters(
            gs, context, rshelmax=rshelmax, rshewmin=rshewmin)
        fxc_calculator = AdiabaticFXCCalculator(localft_calc)
        fxc_kernel = fxc_calculator(functional, '00', qpd)
        Kxc_GG = fxc_kernel.get_Kxc_GG()

        if qpd.kd.gamma:
            Kxc_GG[0, :] = 0.0
            Kxc_GG[:, 0] = 0.0
        Kxc_sGG = np.array([Kxc_GG])
    elif functional[:2] == 'LR':
        p('Calculating LR kernel with alpha = %s' % functional[2:])
        Kxc_sGG = calculate_lr_kernel(qpd, alpha=float(functional[2:]))
    elif functional == 'Bootstrap':
        p('Calculating Bootstrap kernel')
        Kxc_sGG = get_bootstrap_kernel(qpd, chi0_wGG, context)
    else:
        raise ValueError('Invalid functional for the density-density '
                         'xc kernel:', functional)

    return Kxc_sGG[0]


def calculate_lr_kernel(qpd, alpha=0.2):
    """Long range kernel: fxc = \alpha / |q+G|^2"""

    assert qpd.kd.gamma

    f_G = np.zeros(len(qpd.G2_qG[0]))
    f_G[0] = -alpha
    f_G[1:] = -alpha / qpd.G2_qG[0][1:]

    return np.array([np.diag(f_G)])


def get_bootstrap_kernel(qpd, chi0_wGG, context):
    """ Bootstrap kernel (see below) """

    if context.comm.rank == 0:
        chi0_GG = chi0_wGG[0]
        if context.comm.size > 1:
            # If size == 1, chi0_GG is not contiguous, and broadcast()
            # will fail in debug mode.  So we skip it until someone
            # takes a closer look.
            context.comm.broadcast(chi0_GG, 0)
    else:
        nG = qpd.ngmax
        chi0_GG = np.zeros((nG, nG), complex)
        context.comm.broadcast(chi0_GG, 0)

    return calculate_bootstrap_kernel(qpd, chi0_GG, context)


def calculate_bootstrap_kernel(qpd, chi0_GG, context):
    """Bootstrap kernel PRL 107, 186401"""
    p = context.print

    if qpd.kd.gamma:
        v_G = np.zeros(len(qpd.G2_qG[0]))
        v_G[0] = 4 * np.pi
        v_G[1:] = 4 * np.pi / qpd.G2_qG[0][1:]
    else:
        v_G = 4 * np.pi / qpd.G2_qG[0]

    nG = len(v_G)
    K_GG = np.diag(v_G)

    Kxc_GG = np.zeros((nG, nG), dtype=complex)
    dminv_GG = np.zeros((nG, nG), dtype=complex)

    for iscf in range(120):
        dminvold_GG = dminv_GG.copy()
        Kxc_GG = K_GG + Kxc_GG

        chi_GG = np.dot(np.linalg.inv(np.eye(nG, nG)
                                      - np.dot(chi0_GG, Kxc_GG)), chi0_GG)
        dminv_GG = np.eye(nG, nG) + np.dot(K_GG, chi_GG)

        alpha = dminv_GG[0, 0] / (K_GG[0, 0] * chi0_GG[0, 0])
        Kxc_GG = alpha * K_GG
        p(iscf, 'alpha =', alpha, flush=False)
        error = np.abs(dminvold_GG - dminv_GG).sum()
        if np.sum(error) < 0.1:
            p('Self consistent fxc finished in %d iterations !' % iscf)
            break
        if iscf > 100:
            p('Too many fxc scf steps !')

    return np.array([Kxc_GG])



import pytest
import numpy as np
from ase import Atoms

from gpaw import GPAW
from gpaw.tddft import TDDFT, DipoleMomentWriter
from gpaw.inducedfield.inducedfield_base import BaseInducedField
from gpaw.inducedfield.inducedfield_tddft import TDDFTInducedField
from gpaw.poisson import PoissonSolver
from gpaw.test import equal


do_print_values = False  # Use this for printing the reference values

if do_print_values:
    i = 1

    def equal(x, y, tol):  # noqa
        global i
        print("equal(val%d, %20.12f, tol)" % (i, x))
        i += 1


@pytest.mark.later
@pytest.mark.ci
def test_inducedfield_td(in_tmp_dir):
    poisson_eps = 1e-12
    density_eps = 1e-6

    # PoissonSolver
    poissonsolver = PoissonSolver('fd', eps=poisson_eps)

    # Na2 cluster
    atoms = Atoms(symbols='Na2',
                  positions=[(0, 0, 0), (3.0, 0, 0)],
                  pbc=False)
    atoms.center(vacuum=3.0)

    # Standard ground state calculation
    calc = GPAW(mode='fd', nbands=2,
                h=0.6,
                setups={'Na': '1'},
                poissonsolver=poissonsolver,
                symmetry={'point_group': False},
                convergence={'density': density_eps})
    atoms.calc = calc
    _ = atoms.get_potential_energy()
    calc.write('na2_gs.gpw', mode='all')

    # Standard time-propagation initialization
    time_step = 10.0
    iterations = 20
    kick_strength = [1.0e-3, 1.0e-3, 0.0]
    td_calc = TDDFT('na2_gs.gpw')
    DipoleMomentWriter(td_calc, 'na2_td_dm.dat')

    # Create and attach InducedField object
    frequencies = [1.0, 2.08]     # Frequencies of interest in eV
    folding = 'Gauss'             # Folding function
    width = 0.1                   # Line width for folding in eV
    ind = TDDFTInducedField(paw=td_calc,
                            frequencies=frequencies,
                            folding=folding,
                            width=width,
                            restart_file='na2_td.ind')

    # Propagate as usual
    td_calc.absorption_kick(kick_strength=kick_strength)
    td_calc.propagate(time_step, iterations // 2)

    # Save TDDFT and InducedField objects
    td_calc.write('na2_td.gpw', mode='all')
    ind.write('na2_td.ind')
    ind.paw = None

    # Restart and continue
    td_calc = TDDFT('na2_td.gpw')
    DipoleMomentWriter(td_calc, 'na2_td_dm.dat')

    # Load and attach InducedField object
    ind = TDDFTInducedField(filename='na2_td.ind',
                            paw=td_calc,
                            restart_file='na2_td.ind')

    # Continue propagation as usual
    td_calc.propagate(time_step, iterations // 2)

    # Calculate induced electric field
    ind.calculate_induced_field(gridrefinement=2,
                                from_density='comp',
                                poissonsolver=poissonsolver,
                                extend_N_cd=3 * np.ones((3, 2), int),
                                deextend=True)

    # Save
    ind.write('na2_td_field.ind', 'all')
    ind.paw = None
    td_calc = None
    ind = None

    # Read data (test also field data I/O)
    ind = BaseInducedField(filename='na2_td_field.ind',
                           readmode='field')

    # Estimate tolerance (worst case error accumulation)
    tol = (iterations * ind.fieldgd.integrate(ind.fieldgd.zeros() + 1.0) *
           max(density_eps, np.sqrt(poisson_eps)))
    # tol = 0.038905993684
    if do_print_values:
        print('tol = %.12f' % tol)

    # Test
    val1 = ind.fieldgd.integrate(ind.Ffe_wg[0])
    val2 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[0][0]))
    val3 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[0][1]))
    val4 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[0][2]))
    val5 = ind.fieldgd.integrate(ind.Ffe_wg[1])
    val6 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[1][0]))
    val7 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[1][1]))
    val8 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[1][2]))

    equal(val1, 1926.232999117403, tol)
    equal(val2, 0.427606450419, tol)
    equal(val3, 0.565823985683, tol)
    equal(val4, 0.372493489423, tol)
    equal(val5, 1945.618902611449, tol)
    equal(val6, 0.423899965987, tol)
    equal(val7, 0.560882533828, tol)
    equal(val8, 0.369203021329, tol)



from gpaw.grid_descriptor import GridDescriptor
from gpaw.external import (ConstantElectricField, PointChargePotential,
                           PotentialCollection, create_external_potential)


def test_collection():
    a = 4.0
    N = 48
    gd = GridDescriptor((N, N, N), (a, a, a), 0)

    ext1 = ConstantElectricField(1)
    ext2 = PointChargePotential([1, -5], positions=((0, 0, -10), (0, 0, 10)))
    collection = PotentialCollection([ext1, ext2])
    
    ext1.calculate_potential(gd)
    ext2.calculate_potential(gd)
    collection.calculate_potential(gd)

    assert (collection.vext_g == ext1.vext_g + ext2.vext_g).all()

    assert len(collection.todict()['potentials']) == 2
    for text in ['Constant electric', 'Point-charge',
                 collection.__class__.__name__]:
        assert text in collection.__str__()

    # reconstruct
    collection2 = create_external_potential(**collection.todict())
    collection2.calculate_potential(gd)
    assert (collection2.vext_g == collection.vext_g).all()



from ase import __version__

if [int(x) for x in __version__.split('.')[:2]] > [3, 22]:
    from ase.utils.plugins import ExternalIOFormat

    gpaw_yaml = ExternalIOFormat(
        desc='GPAW-yaml output',
        code='+B',
        module='gpaw.yml',
        magic=b'#  __  _  _')



import pytest
from ase import Atoms
from ase.lattice.hexagonal import Hexagonal
from gpaw import GPAW, FermiDirac
from gpaw.response.g0w0 import G0W0


@pytest.fixture
def gpwfile(in_tmp_dir):
    calc = GPAW(
        mode='pw',
        xc='PBE',
        experimental={'niter_fixdensity': 2},
        nbands=16,
        convergence={'bands': 15},
        setups={'Mo': '6'},
        occupations=FermiDirac(0.001),
        kpts={'size': (3, 3, 1), 'gamma': True})

    a = 3.1604
    c = 10.0

    cell = Hexagonal(symbol='Mo',
                     latticeconstant={'a': a, 'c': c}).get_cell()
    layer = Atoms(symbols='MoS2', cell=cell, pbc=[True, True, False],
                  scaled_positions=[(0, 0, 0.5),
                                    (2 / 3, 1 / 3, 0.3 + 0.5),
                                    (2 / 3, 1 / 3, -0.3 + 0.5)])

    pos = layer.get_positions()
    pos[1][2] = pos[0][2] + 3.172 / 2
    pos[2][2] = pos[0][2] - 3.172 / 2
    layer.set_positions(pos)
    layer.calc = calc
    layer.get_potential_energy()
    fname = 'MoS2.gpw'
    calc.write(fname, mode='all')
    return fname


@pytest.mark.response
def test_response_gw_MoS2_cut(scalapack, gpwfile, needs_ase_master):
    gw = G0W0(gpwfile,
              'gw-test',
              nbands=15,
              ecut=10,
              eta=0.2,
              frequencies={'type': 'nonlinear', 'domega0': 0.1},
              truncation='2D',
              kpts=[((1 / 3, 1 / 3, 0))],
              bands=(8, 10))

    e_qp = gw.calculate()['qp'][0, 0]

    paths = gw.savepckl()
    for path in paths.values():
        assert path.exists()

    ev = 2.392
    ec = 7.337
    assert e_qp[0] == pytest.approx(ev, abs=0.01)
    assert e_qp[1] == pytest.approx(ec, abs=0.01)



import numpy as np
import pytest
from ase.build import molecule
from gpaw import GPAW
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.wfwriter import WaveFunctionWriter
from gpaw.mpi import world
from gpaw.tddft.spectrum import photoabsorption_spectrum
from gpaw.test import equal


@pytest.mark.rttddft
def test_lcaotddft_replay(in_tmp_dir):
    atoms = molecule('Na2')
    atoms.center(vacuum=4.0)

    # Ground-state calculation
    calc = GPAW(nbands=2, h=0.4, setups=dict(Na='1'),
                basis='dzp', mode='lcao',
                convergence={'density': 1e-8},
                symmetry={'point_group': False},
                txt='gs.out')
    atoms.calc = calc
    _ = atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    # Time-propagation calculation
    td_calc = LCAOTDDFT('gs.gpw', txt='td.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    WaveFunctionWriter(td_calc, 'wf.ulm')
    WaveFunctionWriter(td_calc, 'wf_split.ulm', split=True)
    td_calc.absorption_kick(np.ones(3) * 1e-5)
    td_calc.propagate(20, 3)
    td_calc.write('td.gpw', mode='all')
    td_calc.propagate(7, 3)

    # Restart from the restart point
    td_calc = LCAOTDDFT('td.gpw', txt='td2.out')
    DipoleMomentWriter(td_calc, 'dm.dat')
    WaveFunctionWriter(td_calc, 'wf.ulm')
    WaveFunctionWriter(td_calc, 'wf_split.ulm')
    td_calc.propagate(20, 3)
    td_calc.propagate(20, 3)
    td_calc.propagate(10, 3)
    photoabsorption_spectrum('dm.dat', 'spec.dat')

    world.barrier()
    ref_i = np.loadtxt('spec.dat').ravel()

    # Replay both wf*.ulm files
    for tag in ['', '_split']:
        td_calc = LCAOTDDFT('gs.gpw', txt='rep%s.out' % tag)
        DipoleMomentWriter(td_calc, 'dm_rep%s.dat' % tag)
        td_calc.replay(name='wf%s.ulm' % tag, update='density')
        photoabsorption_spectrum('dm_rep%s.dat' % tag, 'spec_rep%s.dat' % tag)

        world.barrier()

        # Check the spectrum files
        # Do this instead of dipolemoment files in order to see that the kick
        # was also written correctly in replaying
        data_i = np.loadtxt('spec_rep%s.dat' % tag).ravel()

        tol = 1e-10
        equal(data_i, ref_i, tol)



import numpy as np

from gpaw.mpi import world, broadcast_float
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.dipolemomentwriter import DipoleMomentWriter
from gpaw.lcaotddft.wfwriter import WaveFunctionWriter, WaveFunctionReader
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.frequencydensitymatrix import FrequencyDensityMatrix
from gpaw.tddft.folding import frequencies
from gpaw.utilities import compiled_with_sl


def parallel_options(*, include_kpt=False, fix_sl_auto=False):
    """Generate different parallelization options"""
    parallel_i = []
    for sl_auto in [False, True]:
        if not compiled_with_sl() and sl_auto:
            continue
        for band in [1, 2]:
            for kpt in [1, 2] if include_kpt else [1]:
                if world.size < band * kpt:
                    continue
                parallel = {'sl_auto': sl_auto, 'band': band, 'kpt': kpt}

                if fix_sl_auto and world.size == 1 and parallel['sl_auto']:
                    # Choose BLACS grid manually as the one given by sl_auto
                    # doesn't work well for the small test system and 1 process
                    del parallel['sl_auto']
                    parallel['sl_default'] = (1, 1, 8)

                parallel_i.append(parallel)
    return parallel_i


def calculate_time_propagation(gs_fpath, *, kick,
                               communicator=world, parallel={},
                               do_fdm=False):
    td_calc = LCAOTDDFT(gs_fpath,
                        communicator=communicator,
                        parallel=parallel,
                        txt='td.out')
    if do_fdm:
        dmat = DensityMatrix(td_calc)
        ffreqs = frequencies(range(0, 31, 5), 'Gauss', 0.1)
        fdm = FrequencyDensityMatrix(td_calc, dmat, frequencies=ffreqs)
    DipoleMomentWriter(td_calc, 'dm.dat')
    WaveFunctionWriter(td_calc, 'wf.ulm')
    td_calc.absorption_kick(kick)
    td_calc.propagate(20, 3)
    if do_fdm:
        fdm.write('fdm.ulm')

    communicator.barrier()

    if do_fdm:
        return fdm


def calculate_error(a, ref_a):
    if world.rank == 0:
        err = np.abs(a - ref_a).max()
        print()
        print('ERR', err)
    else:
        err = np.nan
    err = broadcast_float(err, world)
    return err


def check_txt_data(ref_fpath, data_fpath, atol):
    world.barrier()
    ref = np.loadtxt(ref_fpath, encoding='utf-8')
    data = np.loadtxt(data_fpath, encoding='utf-8')
    err = calculate_error(data, ref)
    assert err < atol


def check_wfs(wf_ref_fpath, wf_fpath, atol=1e-12):
    wfr_ref = WaveFunctionReader(wf_ref_fpath)
    wfr = WaveFunctionReader(wf_fpath)
    assert len(wfr) == len(wfr_ref)
    for i in range(1, len(wfr)):
        ref = wfr_ref[i].wave_functions.coefficients
        coeff = wfr[i].wave_functions.coefficients
        err = calculate_error(coeff, ref)
        assert err < atol, f'error at i={i}'


def copy_and_cut_file(src, dst, *, cut_lines=0):
    with open(src, 'r', encoding='utf-8') as fd:
        lines = fd.readlines()
        if cut_lines > 0:
            lines = lines[:-cut_lines]

    with open(dst, 'w', encoding='utf-8') as fd:
        for line in lines:
            fd.write(line)



import numpy as np


class DirectLCAO(object):
    """Eigensolver for LCAO-basis calculation"""

    def __init__(self, diagonalizer=None):
        self.diagonalizer = diagonalizer
        # ??? why should we be able to set
        # this diagonalizer in both constructor and initialize?
        self.has_initialized = False  # XXX

    def initialize(self, gd, dtype, nao, diagonalizer=None):
        self.gd = gd
        self.nao = nao
        if diagonalizer is not None:
            self.diagonalizer = diagonalizer
        assert self.diagonalizer is not None
        self.has_initialized = True  # XXX

    def reset(self):
        pass

    @property
    def error(self):
        return 0.0

    @error.setter
    def error(self, e):
        pass

    def calculate_hamiltonian_matrix(self, hamiltonian, wfs, kpt, Vt_xMM=None,
                                     root=-1, add_kinetic=True):
        # XXX document parallel stuff, particularly root parameter
        assert self.has_initialized

        bfs = wfs.basis_functions

        # distributed_atomic_correction works with ScaLAPACK/BLACS in general.
        # If SL is not enabled, it will not work with band parallelization.
        # But no one would want that for a practical calculation anyway.
        # dH_asp = wfs.atomic_correction.redistribute(wfs, hamiltonian.dH_asp)
        # XXXXX fix atomic corrections
        dH_asp = hamiltonian.dH_asp

        if Vt_xMM is None:
            wfs.timer.start('Potential matrix')
            vt_G = hamiltonian.vt_sG[kpt.s]
            Vt_xMM = bfs.calculate_potential_matrices(vt_G)
            wfs.timer.stop('Potential matrix')

        if bfs.gamma and wfs.dtype == float:
            yy = 1.0
            H_MM = Vt_xMM[0]
        else:
            wfs.timer.start('Sum over cells')
            yy = 0.5
            k_c = wfs.kd.ibzk_qc[kpt.q]
            H_MM = (0.5 + 0.0j) * Vt_xMM[0]

            H_MM += np.einsum('x,xMN->MN',
                              np.exp(2j * np.pi * bfs.sdisp_xc[1:] @ k_c),
                              Vt_xMM[1:],
                              optimize=True)
            wfs.timer.stop('Sum over cells')

        # Add atomic contribution
        #
        #           --   a     a  a*
        # H      += >   P    dH  P
        #  mu nu    --   mu i  ij nu j
        #           aij
        #
        name = wfs.atomic_correction.__class__.__name__
        wfs.timer.start(name)
        wfs.atomic_correction.calculate_hamiltonian(kpt, dH_asp, H_MM, yy)
        wfs.timer.stop(name)

        wfs.timer.start('Distribute overlap matrix')
        H_MM = wfs.ksl.distribute_overlap_matrix(
            H_MM, root, add_hermitian_conjugate=(yy == 0.5))
        wfs.timer.stop('Distribute overlap matrix')

        if add_kinetic:
            H_MM += wfs.T_qMM[kpt.q]

        return H_MM

    def iterate(self, hamiltonian, wfs, occ=None):
        wfs.timer.start('LCAO eigensolver')

        s = -1
        for kpt in wfs.kpt_u:
            if kpt.s != s:
                s = kpt.s
                wfs.timer.start('Potential matrix')
                Vt_xMM = wfs.basis_functions.calculate_potential_matrices(
                    hamiltonian.vt_sG[s])
                wfs.timer.stop('Potential matrix')
            self.iterate_one_k_point(hamiltonian, wfs, kpt, Vt_xMM)
        wfs.set_orthonormalized(True)
        wfs.timer.stop('LCAO eigensolver')

    def iterate_one_k_point(self, hamiltonian, wfs, kpt, Vt_xMM):
        if wfs.bd.comm.size > 1 and wfs.bd.strided:
            raise NotImplementedError

        H_MM = self.calculate_hamiltonian_matrix(hamiltonian, wfs, kpt, Vt_xMM,
                                                 root=0)

        # Decomposition step of overlap matrix can be skipped if we have
        # cached the result and if the solver supports it (Elpa)
        may_decompose = self.diagonalizer.accepts_decomposed_overlap_matrix
        if may_decompose:
            S_MM = wfs.decomposed_S_qMM[kpt.q]
            is_already_decomposed = (S_MM is not None)
            if S_MM is None:
                # Contents of S_MM will be overwritten by eigensolver below.
                S_MM = wfs.decomposed_S_qMM[kpt.q] = wfs.S_qMM[kpt.q].copy()
        else:
            is_already_decomposed = False
            S_MM = wfs.S_qMM[kpt.q]

        if kpt.eps_n is None:
            kpt.eps_n = np.empty(wfs.bd.mynbands)

        diagonalization_string = repr(self.diagonalizer)
        wfs.timer.start(diagonalization_string)
        # May overwrite S_MM (so the results will be stored as decomposed)
        if kpt.C_nM is None:
            kpt.C_nM = wfs.bd.empty(wfs.setups.nao, dtype=wfs.dtype)

        self.diagonalizer.diagonalize(H_MM, kpt.C_nM, kpt.eps_n, S_MM,
                                      is_already_decomposed)
        wfs.timer.stop(diagonalization_string)

        wfs.timer.start('Calculate projections')
        # P_ani are not strictly necessary as required quantities can be
        # evaluated directly using P_aMi/Paaqim.  We should perhaps get rid
        # of the places in the LCAO code using P_ani directly
        wfs.atomic_correction.calculate_projections(wfs, kpt)
        wfs.timer.stop('Calculate projections')

    def __repr__(self):
        # The "diagonalizer" must be equal to the Kohn-Sham layout
        # object presently.  That information will be printed in the
        # text output anyway so we do not need it here.
        #
        # Although maybe it may be better to print it anyway...
        return 'LCAO using direct dense diagonalizer'

    def estimate_memory(self, mem, dtype):
        pass
        # self.diagonalizer.estimate_memory(mem, dtype) #XXX enable this



from ase import Atoms
from gpaw import GPAW
import numpy as np

# spin paired H2


def test_xc_pplda():
    d = 0.75
    h2 = Atoms('H2', [[0, 0, 0], [0, 0, d]])
    h2.center(vacuum=2.)

    e = np.array([])
    f = np.array([])

    for xc in ['LDA', 'PPLDA']:
        calc = GPAW(nbands=-1,
                    xc=xc,
                    convergence={'eigenstates': 1.e-9},
                    txt=None,
                    mode='lcao',
                    basis='szp(dzp)',
                    h=0.3)
        h2.calc = calc
        e = np.append(e, h2.get_potential_energy())
        f = np.append(f, h2.get_forces())
        del calc

    print(e[0], e[1], np.abs(e[0] - e[1]))
    print(f[0], f[1], np.sum(np.abs(f[0] - f[1])))
    assert np.abs(e[0] - e[1]) < 1.e-8
    assert np.sum(np.abs(f[0] - f[1])) < 1.e-8

    # spin polarized O2
    d = 1.2
    o2 = Atoms('O2', [[0, 0, 0], [0, 0, d]], magmoms=[1., 1.])
    o2.center(vacuum=2.)

    e = np.array([])
    f = np.array([])

    for xc in ['LDA', 'PPLDA']:
        calc = GPAW(nbands=-2, xc=xc, convergence={'energy': 1.e-9}, txt=None,
                    mode='lcao', basis='szp(dzp)', h=0.3)
        o2.calc = calc
        e = np.append(e, o2.get_potential_energy())
        f = np.append(f, o2.get_forces())
        del calc

    print(e[0], e[1], np.abs(e[0] - e[1]))
    print(f[0], f[1], np.sum(np.abs(f[0] - f[1])))
    assert np.abs(e[0] - e[1]) < 1.e-8
    assert np.sum(np.abs(f[0] - f[1])) < 1.e-4



from __future__ import annotations

import numpy as np
from ase.units import Bohr, Ha

from gpaw.core.arrays import DistributedArrays as XArray
from gpaw.core.atom_arrays import AtomArrays, AtomDistribution
from gpaw.core.domain import Domain as XDesc
from gpaw.core.uniform_grid import UGArray, UGDesc
from gpaw.mpi import MPIComm
from gpaw.new import zips


class Potential:
    def __init__(self,
                 vt_sR: UGArray,
                 dH_asii: AtomArrays,
                 dedtaut_sR: UGArray | None,
                 energies: dict[str, float],
                 vHt_x: XArray | None = None):
        self.vt_sR = vt_sR
        self.dH_asii = dH_asii
        self.dedtaut_sR = dedtaut_sR
        self.energies = energies
        self.vHt_x = vHt_x  # initial guess for Hartree potential

    def __repr__(self):
        return (f'Potential({self.vt_sR}, {self.dH_asii}, '
                f'{self.dedtaut_sR}, {self.energies})')

    def __str__(self) -> str:
        return (f'potential:\n'
                f'  grid points: {self.vt_sR.desc.size}\n')

    def dH(self, P_ani, out_ani, spin):
        if len(P_ani.dims) == 1:  # collinear wave functions
            P_ani.block_diag_multiply(self.dH_asii, out_ani, spin)
            return

        # Non-collinear wave functions:
        P_ansi = P_ani
        out_ansi = out_ani

        for (a, P_nsi), out_nsi in zips(P_ansi.items(), out_ansi.values()):
            v_ii, x_ii, y_ii, z_ii = (dh_ii.T for dh_ii in self.dH_asii[a])
            assert v_ii.dtype == complex
            out_nsi[:, 0] = (P_nsi[:, 0] @ (v_ii + z_ii) +
                             P_nsi[:, 1] @ (x_ii - 1j * y_ii))
            out_nsi[:, 1] = (P_nsi[:, 1] @ (v_ii - z_ii) +
                             P_nsi[:, 0] @ (x_ii + 1j * y_ii))
        return out_ansi

    def move(self, atomdist: AtomDistribution) -> None:
        """Move atoms inplace."""
        self.dH_asii = self.dH_asii.moved(atomdist)

    def redist(self,
               grid: UGDesc,
               desc: XDesc,
               atomdist: AtomDistribution,
               comm1: MPIComm,
               comm2: MPIComm) -> Potential:
        return Potential(
            self.vt_sR.redist(grid, comm1, comm2),
            self.dH_asii.redist(atomdist, comm1, comm2),
            None if self.dedtaut_sR is None else self.dedtaut_sR.redist(
                grid, comm1, comm2),
            self.energies.copy(),
            None if self.vHt_x is None else self.vHt_x.redist(
                desc, comm1, comm2))

    def _write_gpw(self, writer, ibzwfs):
        from gpaw.new.calculation import combine_energies
        energies = combine_energies(self, ibzwfs)
        energies['band'] = ibzwfs.energies['band']
        if 'stress' in self.energies:
            energies['stress'] = self.energies['stress']
        dH_asp = self.dH_asii.to_cpu().to_lower_triangle().gather()
        vt_sR = self.vt_sR.to_xp(np).gather()
        assert self.vHt_x is not None
        vHt_x = self.vHt_x.to_xp(np).gather()
        if self.dedtaut_sR is not None:
            dedtaut_sR = self.dedtaut_sR.to_xp(np).gather()
        if dH_asp is None:
            return
        writer.write(
            potential=vt_sR.data * Ha,
            electrostatic_potential=vHt_x.data * Ha,
            atomic_hamiltonian_matrices=dH_asp.data * Ha,
            **{f'e_{name}': val * Ha for name, val in energies.items()})
        if self.dedtaut_sR is not None:
            writer.write(mgga_potential=dedtaut_sR.data * Bohr**3)



# Written by Lauri Lehtovaara 2008
import numpy as np

from gpaw.utilities.blas import axpy


class MultiBlas:
    def __init__(self, gd):
        self.gd = gd

    # Multivector ZAXPY: a x + y => y
    def multi_zaxpy(self, a, x, y, nvec):
        if isinstance(a, (float, complex)):
            for i in range(nvec):
                axpy(a * (1 + 0j), x[i], y[i])
        else:
            for i in range(nvec):
                axpy(a[i] * (1.0 + 0.0j), x[i], y[i])

    # Multivector dot product, a^H b, where ^H is transpose
    def multi_zdotc(self, s, x, y, nvec):
        for i in range(nvec):
            s[i] = np.vdot(x[i], y[i])
        self.gd.comm.sum(s)
        return s

    # Multiscale: a x => x
    def multi_scale(self, a, x, nvec):
        if isinstance(a, (float, complex)):
            x *= a
        else:
            for i in range(nvec):
                x[i] *= a[i]


class BandPropertyMonitor:
    def __init__(self, wfs, name, interval=1):
        self.niter = 0
        self.interval = interval

        self.wfs = wfs

        self.name = name

    def __call__(self):
        self.update(self.wfs)
        self.niter += self.interval

    def update(self, wfs):
        # strictly serial XXX!
        data_un = []

        for u, kpt in enumerate(wfs.kpt_u):
            data_n = getattr(kpt, self.name)

            data_un.append(data_n)

        self.write(np.array(data_un))

    def write(self, data):
        pass


class BandPropertyWriter(BandPropertyMonitor):
    def __init__(self, filename, wfs, name, interval=1):
        BandPropertyMonitor.__init__(self, wfs, name, interval)
        self.fileobj = open(filename, 'w')

    def write(self, data):
        self.fileobj.write(data.tostring())
        self.fileobj.flush()

    def __del__(self):
        self.fileobj.close()


class StaticOverlapMonitor:
    def __init__(self, wfs, wf_u, P_aui, interval=1):
        self.niter = 0
        self.interval = interval

        self.wfs = wfs

        self.wf_u = wf_u
        self.P_aui = P_aui

    def __call__(self):
        self.update(self.wfs)
        self.niter += self.interval

    def update(self, wfs, calculate_P_ani=False):
        # strictly serial XXX!
        Porb_un = []

        for u, kpt in enumerate(wfs.kpt_u):
            swf = self.wf_u[u].ravel()

            psit_n = kpt.psit_nG.reshape((len(kpt.f_n), -1))
            Porb_n = np.dot(psit_n.conj(), swf) * wfs.gd.dv

            P_ani = kpt.P_ani

            if calculate_P_ani:
                # wfs.pt.integrate(psit_nG, P_ani, kpt.q)
                raise NotImplementedError(
                    'In case you were wondering, TODO XXX')

            for a, P_ni in P_ani.items():
                sP_i = self.P_aui[a][u]
                for n in range(wfs.bd.nbands):
                    for i in range(len(P_ni[0])):
                        for j in range(len(P_ni[0])):
                            Porb_n[n] += (P_ni[n][i].conj() *
                                          wfs.setups[a].dO_ii[i][j] *
                                          sP_i[j])

            Porb_un.append(Porb_n)

        self.write(np.array(Porb_un))

    def write(self, data):
        pass


class StaticOverlapWriter(StaticOverlapMonitor):
    def __init__(self, filename, wfs, overlap, interval=1):
        StaticOverlapMonitor.__init__(self, wfs, overlap, interval)
        self.fileobj = open(filename, 'w')

    def write(self, data):
        self.fileobj.write(data.tostring())
        self.fileobj.flush()

    def __del__(self):
        self.fileobj.close()


class DynamicOverlapMonitor:
    def __init__(self, wfs, overlap, interval=1):
        self.niter = 0
        self.interval = interval

        self.setups = overlap.setups
        self.operator = overlap.operator
        self.wfs = wfs

    def __call__(self):
        self.update(self.wfs)
        self.niter += self.interval

    def update(self, wfs, calculate_P_ani=False):

        # strictly serial XXX!
        S_unn = []

        for kpt in wfs.kpt_u:
            psit_nG = kpt.psit_nG
            P_ani = kpt.P_ani

            if calculate_P_ani:
                # wfs.pt.integrate(psit_nG, P_ani, kpt.q)
                raise NotImplementedError(
                    'In case you were wondering, TODO XXX')

            # Construct the overlap matrix:
            def S(x):
                return x

            dS_aii = dict([(a, self.setups[a].dO_ii) for a in P_ani])
            S_nn = self.operator.calculate_matrix_elements(psit_nG, P_ani,
                                                           S, dS_aii)
            S_unn.append(S_nn)

        self.write(np.array(S_unn))

    def write(self, data):
        pass


class DynamicOverlapWriter(DynamicOverlapMonitor):
    def __init__(self, filename, wfs, overlap, interval=1):
        DynamicOverlapMonitor.__init__(self, wfs, overlap, interval)
        self.fileobj = open(filename, 'w')

    def write(self, data):
        self.fileobj.write(data.tostring())
        self.fileobj.flush()

    def __del__(self):
        self.fileobj.close()



import numpy as np
import pytest

import gpaw.mpi as mpi
from gpaw.grid_descriptor import GridDescriptor
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.lfc import PWLFC
from gpaw.spline import Spline
from gpaw.test import equal


@pytest.mark.ci
def test_pw_lfc():
    x = 2.0
    rc = 3.5
    r = np.linspace(0, rc, 100)

    n = 40
    a = 8.0
    gd = GridDescriptor((n, n, n), (a, a, a), comm=mpi.serial_comm)

    kpts = np.array([(0.25, 0.25, 0.0)])
    kd = KPointDescriptor(kpts)
    spos_ac = np.array([(0.15, 0.5, 0.95)])

    pd = PWDescriptor(45, gd, complex, kd)

    eikr = np.ascontiguousarray(
        np.exp(2j * np.pi * np.dot(np.indices(gd.N_c).T,
                                   (kpts / gd.N_c).T).T)[0])

    for l in range(3):
        print(l)
        s = Spline(l, rc, 2 * x**1.5 / np.pi * np.exp(-x * r**2))

        lfc1 = LFC(gd, [[s]], kd, dtype=complex)
        lfc2 = PWLFC([[s]], pd)

        c_axi = {0: np.zeros((1, 2 * l + 1), complex)}
        c_axi[0][0, 0] = 1.9 - 4.5j
        c_axiv = {0: np.zeros((1, 2 * l + 1, 3), complex)}

        b1 = gd.zeros(1, dtype=complex)
        b2 = pd.zeros(1, dtype=complex)

        for lfc, b in [(lfc1, b1), (lfc2, b2)]:
            lfc.set_positions(spos_ac)
            lfc.add(b, c_axi, 0)

        b2 = pd.ifft(b2[0]) * eikr
        equal(abs(b2 - b1[0]).max(), 0, 0.001)

        b1 = eikr[None]
        b2 = pd.fft(b1[0] * 0 + 1).reshape((1, -1))

        results = []
        results2 = []
        for lfc, b in [(lfc1, b1), (lfc2, b2)]:
            lfc.integrate(b, c_axi, 0)
            results.append(c_axi[0][0].copy())
            lfc.derivative(b, c_axiv, 0)
            results2.append(c_axiv[0][0].copy())
        equal(abs(np.ptp(results2, 0)).max(), 0, 1e-7)
        equal(abs(np.ptp(results, 0)).max(), 0, 3e-8)



import numpy as np
from scipy.special import spherical_jn

from gpaw.ffbt import rescaled_fourier_bessel_transform
from gpaw.gaunt import gaunt, super_gaunt
from gpaw.spherical_harmonics import Y
from gpaw.sphere.rshe import RealSphericalHarmonicsExpansion
from types import SimpleNamespace


class Setuplet:
    def __init__(self, *, phit_jg, phi_jg, rgd, l_j, rcut_j):
        self.rgd = rgd
        self.data = SimpleNamespace(phit_jg=phit_jg, phi_jg=phi_jg)
        self.l_j = l_j
        self.ni = np.sum([2 * l + 1 for l in l_j])
        self.rcut_j = rcut_j


def calculate_pair_density_correction(qG_Gv, *, pawdata):
    r"""Calculate the atom-centered PAW correction to the pair density.
                                                      ˍ
    The atom-centered pair density correction tensor, Q_aii', is defined as the
    atom-centered Fourier transform

    ˍ             /                                     ˷         ˷
    Q_aii'(G+q) = | dr e^-i(G+q)r [φ_ai^*(r) φ_ai'(r) - φ_ai^*(r) φ_ai'(r)]
                  /

    evaluated with the augmentation sphere center at the origin. The full pair
    density correction tensor is then given by
                                  ˍ
    Q_aii'(G+q) = e^(-i[G+q].R_a) Q_aii'(G+q)

    Expanding the plane wave coefficient into real spherical harmonics and
    spherical Bessel functions, the correction can split into angular and
    radial contributions

                    l
                __  __
                \   \      l  m ˰   m,m_i,m_i'
    Q_aii'(K) = /   /  (-i)  Y (K) g
                ‾‾  ‾‾        l     l,l_i,l_i'
                l  m=-l
                            rc
                            /                    a     a      ˷a    ˷a
                       × 4π | r^2 dr j_l(|K|r) [φ (r) φ (r) - φ (r) φ (r)]
                            /                    j_i   j_i'    j_i   j_i'
                            0

    where K = G+q and g denotes the Gaunt coefficients.

    For more information, see [PRB 103, 245110 (2021)]. In particular, it
    should be noted that the partial waves themselves are defined via real
    spherical harmonics and radial functions φ_j from the PAW setups:

     a       m_i ˰   a
    φ (r) = Y   (r) φ (r)
     i       l_i     j_i
    """
    rgd = pawdata.rgd
    ni = pawdata.ni  # Number of partial waves
    l_j = pawdata.l_j  # l-index for each radial function index j
    G_LLL = gaunt(max(l_j))
    # (Real) radial functions for the partial waves
    phi_jg = pawdata.data.phi_jg
    phit_jg = pawdata.data.phit_jg

    # Grid cutoff to create spline representation
    gcut2 = rgd.ceil(2 * max(pawdata.rcut_j))

    # Initialize correction tensor
    npw = qG_Gv.shape[0]
    Qbar_Gii = np.zeros((npw, ni, ni), dtype=complex)

    # K-vector norm
    k_G = np.linalg.norm(qG_Gv, axis=1)

    # Loop of radial function indices for partial waves i and i'
    i1_counter = 0
    for j1, l1 in enumerate(l_j):
        i2_counter = 0
        for j2, l2 in enumerate(l_j):
            # Calculate the radial partial wave correction
            #                              ˷      ˷
            # Δn_jj'(r) = φ_j(r) φ_j'(r) - φ_j(r) φ_j'(r)
            dn_g = phi_jg[j1] * phi_jg[j2] - phit_jg[j1] * phit_jg[j2]

            # Sample l according to the Gaunt coefficient selection rules, see
            # e.g. gpaw.test.test_gaunt
            for l in range(abs(l1 - l2), l1 + l2 + 1, 2):
                # To evaluate the radial integral efficiently, we rely on the
                # Fast Fourier Bessel Transform (FFBT) algorithm, see gpaw.ffbt
                # In order to do so, we make a spline representation of the
                # radial partial wave correction rescaled with a factor of r^-l
                spline = rgd.spline(dn_g[:gcut2], l=l, points=2**12)
                # This allows us to calculate a spline representation of the
                # spherical Fourier-Bessel transform
                #                 rc
                #             4π  /
                # Δn_jj'(k) = ‾‾‾ | r^2 dr j_l(kr) Δn_jj'(r)
                #             k^l /
                #                 0
                kspline = rescaled_fourier_bessel_transform(spline, N=2**14)

                # Now, this implementation relies on a range of hardcoded
                # values, which are not guaranteed to work for all cases.
                # In particular, the uniform radial grid used for the FFBT is
                # defined through the `rcut` and `N` parameters in the
                # `rescaled_fourier_bessel_transform()` function (the former is
                # currently hardcoded), and the `points` parameter to
                # `rgd.spline()` controls an intermediate interpolation step.
                # To make a generally reliable implementation, one would need
                # to control these parameters based on the setup, e.g. the
                # nonlinear radial grid spacing. In doing so, one should be
                # mindful that the rcut parameter defines the reciprocal grid
                # spacing of the kspline representation and that N controls
                # the k-range (which might need to depend on input qG_Gv).

                # For now, we simply check that the requested plane waves are
                # within the computed k-range of the FFBT and check that the
                # resulting transforms match a manual calculation at a few
                # selected K-vectors.
                kmax = np.max(k_G)
                assert kmax <= kspline.get_cutoff()
                # Manual calculation at kmax
                dnmax = rgd.integrate(spherical_jn(l, kmax * rgd.r_g) * dn_g)
                # Manual calculation at average k
                kavg = np.average(k_G)
                dnavg = rgd.integrate(spherical_jn(l, kavg * rgd.r_g) * dn_g)
                k_k = [kmax, kavg]
                dn_k = [dnmax, dnavg]
                if l == 0:
                    # Manual calculation at k=0
                    k_k.append(0.)
                    dn_k.append(rgd.integrate(dn_g))
                k_k = np.array(k_k)
                dn_k = np.array(dn_k)
                assert np.allclose(k_k**l * kspline.map(k_k), dn_k,
                                   rtol=1e-2, atol=1e-3), \
                    f'FFBT mismatch: {k_k**l * kspline.map(k_k)}, {dn_k}'

                # Finally, we can map the Fourier-Bessel transform onto the
                # the requested k-vectors of the input plane wave basis
                dn_G = kspline.map(k_G)

                # Angular part of the integral
                f_G = (-1j)**l * dn_G
                # Generate m-indices for each radial function
                for m1 in range(2 * l1 + 1):
                    for m2 in range(2 * l2 + 1):
                        # Set up the i=(l,m) index for each partial wave
                        i1 = i1_counter + m1
                        i2 = i2_counter + m2
                        # Extract Gaunt coefficients
                        G_m = G_LLL[l1**2 + m1, l2**2 + m2, l**2:(l + 1)**2]
                        for m, gaunt_coeff in enumerate(G_m):
                            if gaunt_coeff == 0:
                                continue
                            # Calculate the solid harmonic
                            #        m ˰
                            # |K|^l Y (K)
                            #        l
                            klY_G = Y(l**2 + m, *qG_Gv.T)
                            # Add contribution to the PAW correction
                            Qbar_Gii[:, i1, i2] += gaunt_coeff * klY_G * f_G

            # Add to i and i' counters
            i2_counter += 2 * l2 + 1
        i1_counter += 2 * l1 + 1
    return Qbar_Gii


def calculate_matrix_element_correction(qG_Gv, pawdata,
                                        rshe: RealSphericalHarmonicsExpansion):
    r"""Calculate the atom-centered correction to a generalized matrix element.

    For matrix elements corresponding to the expectation value of a plane wave
    coefficient e^-i(G+q)r and a known functional of the (spin-)density
    f[n](r), the PAW correction tensor is given by

    F_aii'(G+q) = <φ_ai| e^-i(G+q)r f[n](r) |φ_ai'>
                     ˷                         ˷
                  - <φ_ai| e^-i(G+q)r f[n](r) |φ_ai'>
                                  ˍ
                = e^(-i[G+q].R_a) F_aii'(G+q)
          ˍ
    where F_aii'(G+q) is the atom-centered PAW correction tensor.

    Expanding the functional f[n](r) in the atom-centered frame in real
    spherical harmonics (corresponding to the input rshe),

                  l
              __  __
         →    \   \   m ˰   m
    f[n](r) = /   /  Y (r) f (r)
              ‾‾  ‾‾  l     l
              l  m=-l

    expansion of the plane-wave coefficient in real spherical harmonics and
    spherical Bessel functions j_l(Kr) yields the following expression for the
    atom-centered correction tensor [publication in preparation]:

                       l        l'
                   __  __   __  __
    ˍ              \   \    \   \      l'  m'˰   m_i,m_i',m,m'
    F_aii'(K) = 4π /   /    /   /  (-i)   Y (K) G
                   ‾‾  ‾‾   ‾‾  ‾‾         l'    l_i,l_i',l,l'
                   l  m=-l  l' m'=-l'

                                rc
                                /  2            a     a      ˷a    ˷a      m
                              × | r dr j (Kr) [φ (r) φ (r) - φ (r) φ (r)] f (r)
                                /       l'      j_i   j_i'    j_i   j_i'   l
                                0

    where K=G+q and G_LLLL denotes the super Gaunt coefficients, which yield
    the integrals over four spherical harmonics.
    """
    rgd = rshe.rgd
    assert rgd is pawdata.xc_correction.rgd
    ni = pawdata.ni  # Number of partial waves
    l_j = pawdata.l_j  # l-index for each radial function index j
    lmax = max(l_j)
    assert max(rshe.l_M) <= 2 * lmax
    G_LLLL = super_gaunt(lmax)
    # (Real) radial functions for the partial waves
    phi_jg = pawdata.data.phi_jg
    phit_jg = pawdata.data.phit_jg
    # Truncate the radial functions to span only the radial grid coordinates
    # which need correction
    assert np.allclose(rgd.r_g, pawdata.rgd.r_g[:rgd.N])
    phi_jg = np.array(phi_jg)[:, :rgd.N]
    phit_jg = np.array(phit_jg)[:, :rgd.N]

    # Initialize correction tensor
    npw = qG_Gv.shape[0]
    Fbar_Gii = np.zeros((npw, ni, ni), dtype=complex)

    # K-vector norm and direction
    k_G = np.linalg.norm(qG_Gv, axis=1)
    Kd_Gv = qG_Gv.copy()
    Kd_Gv[k_G > 1e-10] /= k_G[k_G > 1e-10, np.newaxis]

    # Loop of radial function indices for partial waves i and i'
    i1_counter = 0
    for j1, l1 in enumerate(l_j):
        i2_counter = 0
        for j2, l2 in enumerate(l_j):
            # Calculate the radial partial wave correction
            #                              ˷      ˷
            # Δn_jj'(r) = φ_j(r) φ_j'(r) - φ_j(r) φ_j'(r)
            dn_g = phi_jg[j1] * phi_jg[j2] - phit_jg[j1] * phit_jg[j2]

            # Loop through the angular components in the real spherical
            # harmonics expansion of f[n](r)
            for l, L, f_g in zip(rshe.l_M, rshe.L_M, rshe.f_gM.T):
                # Apply Gaunt coefficient selection rules to loop through
                # the l' coefficients of the plane-wave expansion
                lpmin = np.min(abs(
                    np.arange(abs(l1 - l2), l1 + l2 + 1) - l))
                for lp in range(lpmin, l1 + l2 + l + 1):
                    if not (l1 + l2 + l + lp) % 2 == 0:
                        continue
                    # --- Calculate radial part of the correction --- #
                    # Vectorize calculation of spherical Bessel functions
                    lp_Gg = lp * np.ones((npw, rgd.N), dtype=int)
                    kr_Gg = k_G[:, np.newaxis] * rgd.r_g[np.newaxis]
                    jl_Gg = spherical_jn(lp_Gg, kr_Gg)  # so slow...
                    # Integrate correction
                    dnf_G = rgd.integrate_trapz(
                        jl_Gg * dn_g[np.newaxis] * f_g[np.newaxis])

                    # --- Calculate angular part of the correction --- #
                    x_G = 4 * np.pi * (-1j)**lp * dnf_G
                    # Loop through available m-indices for the partial waves
                    # and generate the composite L=(l,m) index as well as the
                    # partial wave index i
                    for m1 in range(2 * l1 + 1):
                        L1 = l1**2 + m1
                        i1 = i1_counter + m1
                        for m2 in range(2 * l2 + 1):
                            L2 = l2**2 + m2
                            i2 = i2_counter + m2
                            # Loop through m' indices of the plane-wave
                            # expansion and generate the L' composite index
                            for mp in range(2 * lp + 1):
                                Lp = lp**2 + mp
                                # If the angular integral (super gaunt
                                # coefficient) is finite,
                                coeff = G_LLLL[L1, L2, L, Lp]
                                if abs(coeff) > 1e-10:
                                    # Calculate spherical harmonic and add
                                    # contribution to the PAW correction
                                    Y_G = Y(Lp, *Kd_Gv.T)
                                    Fbar_Gii[:, i1, i2] += coeff * Y_G * x_G

            # Add to i and i' counters
            i2_counter += 2 * l2 + 1
        i1_counter += 2 * l1 + 1
    return Fbar_Gii


class PWPAWCorrectionData:
    def __init__(self, Q_aGii, qpd, pawdatasets, pos_av, atomrotations):
        # Sometimes we loop over these in ways that are very dangerous.
        # It must be list, not dictionary.
        assert isinstance(Q_aGii, list)
        assert len(Q_aGii) == len(pos_av) == len(pawdatasets)

        self.Q_aGii = Q_aGii

        self.qpd = qpd
        self.pawdatasets = pawdatasets
        self.pos_av = pos_av
        self.atomrotations = atomrotations

    def _new(self, Q_aGii):
        return PWPAWCorrectionData(Q_aGii, qpd=self.qpd,
                                   pawdatasets=self.pawdatasets,
                                   pos_av=self.pos_av,
                                   atomrotations=self.atomrotations)

    def remap(self, M_vv, G_Gv, sym, sign):
        Q_aGii = []
        for a, Q_Gii in enumerate(self.Q_aGii):
            x_G = self._get_x_G(G_Gv, M_vv, self.pos_av[a])
            U_ii = self.atomrotations.get_by_a(a).R_sii[sym]

            Q_Gii = np.einsum('ij,kjl,ml->kim',
                              U_ii,
                              Q_Gii * x_G[:, None, None],
                              U_ii,
                              optimize='optimal')
            if sign == -1:
                Q_Gii = Q_Gii.conj()
            Q_aGii.append(Q_Gii)

        return self._new(Q_aGii)

    def _get_x_G(self, G_Gv, M_vv, pos_v):
        # This doesn't really belong here.  Or does it?  Maybe this formula
        # is only used with PAW corrections.
        return np.exp(1j * (G_Gv @ (pos_v - M_vv @ pos_v)))

    def remap_by_symop(self, symop, G_Gv, M_vv):
        return self.remap(M_vv, G_Gv, symop.symno, symop.sign)

    def multiply(self, P_ani, band):
        assert isinstance(P_ani, list)
        assert len(P_ani) == len(self.Q_aGii)

        C1_aGi = [Qa_Gii @ P1_ni[band].conj()
                  for Qa_Gii, P1_ni in zip(self.Q_aGii, P_ani)]
        return C1_aGi

    def reduce_ecut(self, G2G):
        # XXX actually we should return this with another PW descriptor.
        return self._new([Q_Gii.take(G2G, axis=0) for Q_Gii in self.Q_aGii])

    def almost_equal(self, otherpawcorr, G_G):
        for a, Q_Gii in enumerate(otherpawcorr.Q_aGii):
            e = abs(self.Q_aGii[a] - Q_Gii[G_G]).max()
            if e > 1e-12:
                return False
        return True


def get_pair_density_paw_corrections(pawdatasets, qpd, spos_ac, atomrotations):
    r"""Calculate and bundle paw corrections to the pair densities as a
    PWPAWCorrectionData object.

    The pair density PAW correction tensor is given by:

                  /
    Q_aii'(G+q) = | dr e^(-i[G+q].r) [φ_ai^*(r-R_a) φ_ai'(r-R_a)
                  /                     ˷             ˷
                                      - φ_ai^*(r-R_a) φ_ai'(r-R_a)]
    """
    qG_Gv = qpd.get_reciprocal_vectors(add_q=True)
    pos_av = spos_ac @ qpd.gd.cell_cv

    # Calculate pair density PAW correction tensor
    Q_aGii = []
    for pawdata, pos_v in zip(pawdatasets, pos_av):
        # Calculate atom-centered correction tensor
        Qbar_Gii = calculate_pair_density_correction(qG_Gv, pawdata=pawdata)
        # Add dependency on the atomic position (phase factor)
        x_G = np.exp(-1j * (qG_Gv @ pos_v))
        Q_aGii.append(x_G[:, np.newaxis, np.newaxis] * Qbar_Gii)

    return PWPAWCorrectionData(Q_aGii, qpd=qpd,
                               pawdatasets=pawdatasets,
                               pos_av=pos_av,
                               atomrotations=atomrotations)




import numpy as np
from ase.units import Bohr, _hbar, _e, _me, _eps0
from ase.utils.timing import Timer
from ase.parallel import parprint
from gpaw.mpi import world
from gpaw.nlopt.basic import load_data
from gpaw.nlopt.matrixel import get_rml, get_derivative
from gpaw.utilities.progressbar import ProgressBar


def get_shg(
        freqs=[1.0],
        eta=0.05,
        pol='yyy',
        eshift=0.0,
        gauge='lg',
        ftol=1e-4, Etol=1e-6,
        band_n=None,
        out_name='shg.npy',
        mml_name='mml.npz'):
    """Calculate RPA SHG spectrum for nonmagnetic semiconductors.

    Output: shg.npy file with numpy array containing the spectrum and
    frequencies.

    Parameters:

    freqs:
        Excitation frequency array (a numpy array or list)
    eta:
        Broadening, a number or an array (default 0.05 eV)
    pol:
        Tensor element (default 'yyy')
    gauge:
        Choose the gauge (lg or vg)
    Etol, ftol:
        Tol. in energy and fermi to consider degeneracy
    band_n:
        List of bands in the sum (default 0 to nb)
    out_name:
        Output filename (default 'shg.npy')
    mml_name:
        The momentum filename (default 'mml.npz')
    """

    # Start a timer
    timer = Timer()
    parprint('Calculating SHG spectrum (in {:d} cores).'.format(world.size))

    # Useful variables
    pol_v = ['xyz'.index(ii) for ii in pol]
    freqs = np.array(freqs)
    nw = len(freqs)
    w_lc = freqs + 1e-12 + 1j * eta  # Add small value to avoid 0
    # Use the TRS to reduce calculation time
    w_l = np.hstack((-w_lc[-1::-1], w_lc))
    nw = 2 * nw
    parprint('Calculation in the {} gauge for element {}.'.format(gauge, pol))

    # Load the required data
    with timer('Load and distribute the data'):
        k_info = load_data(mml_name=mml_name)
        if k_info:
            tmp = list(k_info.values())[0]
            nb = len(tmp[1])
            nk = len(k_info) * world.size  # Approximately
            if band_n is None:
                band_n = list(range(nb))
            mem = 6 * 3 * nk * nb**2 * 16 / 2**20
            parprint(f'At least {mem:.2f} MB of memory is required.')

    # Initial call to print 0% progress
    count = 0
    ncount = len(k_info)
    if world.rank == 0:
        pb = ProgressBar()

    # Initialize the outputs
    sum2_l = np.zeros((nw), complex)
    sum3_l = np.zeros((nw), complex)

    # Do the calculations
    for _, (we, f_n, E_n, p_vnn) in k_info.items():
        # Which gauge
        if gauge == 'vg':
            with timer('Sum over bands'):
                tmp = shg_velocity_gauge(
                    w_l, f_n, E_n, p_vnn, pol_v,
                    band_n, ftol, Etol, eshift)
        elif gauge == 'lg':
            with timer('Position matrix elements calculation'):
                r_vnn, D_vnn = get_rml(E_n, p_vnn, pol_v, Etol=Etol)

            with timer('Compute generalized derivative'):
                rd_vvnn = get_derivative(E_n, r_vnn, D_vnn, pol_v, Etol=Etol)

            with timer('Sum over bands'):
                tmp = shg_length_gauge(
                    w_l, f_n, E_n, r_vnn, rd_vvnn, D_vnn, pol_v,
                    band_n, ftol, Etol, eshift)
        else:
            parprint('Gauge ' + gauge + ' not implemented.')
            raise NotImplementedError

        # Add it to previous with a weight
        sum2_l += tmp[0] * we
        sum3_l += tmp[1] * we

        # Print the progress
        if world.rank == 0:
            pb.update(count / ncount)
        count += 1

    if world.rank == 0:
        pb.finish()
    with timer('Gather data from cores'):
        world.sum(sum2_l)
        world.sum(sum3_l)

    # Make the output in SI unit
    chi_l = make_output(gauge, sum2_l, sum3_l)
    # A multi-col output
    nw = len(freqs)
    chi_l = chi_l[nw:] + chi_l[nw - 1::-1]
    shg = np.vstack((freqs, chi_l))

    # Save it to the file
    if world.rank == 0:
        np.save(out_name, shg)

        # Print the timing
        timer.write()

    return shg


def shg_velocity_gauge(
        w_l, f_n, E_n, p_vnn, pol_v,
        band_n=None, ftol=1e-4, Etol=1e-6, eshift=0):
    """
    Loop over bands for computing in velocity gauge

    Input:
        w_l             Complex frequency array
        f_n             Fermi levels
        E_n             Energies
        p_vnn           Momentum matrix elements
        pol_v           Tensor element
        band_n          Band list
        Etol, ftol      Tol. in energy and fermi to consider degeneracy
        eshift          Bandgap correction
    Output:
        sum2_l, sum3_l  Output 2 and 3 bands terms
    """

    # Initialize variables
    nb = len(f_n)
    if band_n is None:
        band_n = list(range(nb))
    sum2_l = np.zeros(w_l.size, complex)
    sum3_l = np.zeros(w_l.size, complex)

    # Loop over bands
    for nni in band_n:
        for mmi in band_n:
            # Remove non important term using TRS
            if mmi <= nni:
                continue

            # Useful variables
            fnm = f_n[nni] - f_n[mmi]
            Emn = E_n[mmi] - E_n[nni] + fnm * eshift

            # Comute the 2-band term
            if np.abs(Emn) > Etol and np.abs(fnm) > ftol:
                pnml = (p_vnn[pol_v[0], nni, mmi]
                        * (p_vnn[pol_v[1], mmi, nni]
                            * (p_vnn[pol_v[2], mmi, mmi]
                                - p_vnn[pol_v[2], nni, nni])
                            + p_vnn[pol_v[2], mmi, nni]
                            * (p_vnn[pol_v[1], mmi, mmi]
                                - p_vnn[pol_v[1], nni, nni])) / 2)
                sum2_l += 1j * fnm * np.imag(pnml) * \
                    (1 / (Emn**4 * (w_l - Emn)) -
                        16 / (Emn**4 * (2 * w_l - Emn)))

            # Loop over the last band index
            for lli in band_n:
                fnl = f_n[nni] - f_n[lli]
                fml = f_n[mmi] - f_n[lli]

                # Do not do zero calculations
                if np.abs(fnl) < ftol and np.abs(fml) < ftol:
                    continue

                # Compute the susceptibility with 1/w form
                Eln = E_n[lli] - E_n[nni] + fnl * eshift
                Eml = E_n[mmi] - E_n[lli] - fml * eshift
                pnml = (p_vnn[pol_v[0], nni, mmi]
                        * (p_vnn[pol_v[1], mmi, lli]
                            * p_vnn[pol_v[2], lli, nni]
                            + p_vnn[pol_v[2], mmi, lli]
                            * p_vnn[pol_v[1], lli, nni]))
                pnml = 1j * np.imag(pnml) / 2

                # Compute the divergence-free terms
                if np.abs(Emn) > Etol and np.abs(
                        Eml) > Etol and np.abs(Eln) > Etol:
                    ftermD = (16 / (Emn**3 * (2 * w_l - Emn))
                              * (fnl / (Emn - 2 * Eln)
                                 + fml / (Emn - 2 * Eml))) \
                        + fnl / (Eln**3 * (2 * Eln - Emn)
                                 * (w_l - Eln)) \
                        + fml / (Eml**3 * (2 * Eml - Emn)
                                 * (w_l - Eml))
                    sum3_l += pnml * ftermD

    return sum2_l, sum3_l


def shg_length_gauge(
        w_l, f_n, E_n, r_vnn, rd_vvnn, D_vnn, pol_v,
        band_n=None, ftol=1e-4, Etol=1e-6, eshift=0):
    """
    Loop over bands for computing in length gauge

    Input:
        w_l             Complex frequency array
        f_n             Fermi levels
        E_n             Energies
        r_vnn           Momentum matrix elements
        rd_vvnn         Generalized derivative of position
        D_vnn           Velocity difference
        pol_v           Tensor element
        band_n          Band list
        Etol, ftol      Tol. in energy and fermi to consider degeneracy
        eshift          Bandgap correction
    Output:
        sum2_l, sum3_l  Output 2 and 3 bands terms
    """

    # Initialize variables
    nb = len(f_n)
    if band_n is None:
        band_n = list(range(nb))
    sum2_l = np.zeros(w_l.size, complex)
    sum3_l = np.zeros(w_l.size, complex)

    # Loop over bands
    for nni in band_n:
        for mmi in band_n:
            # Remove the non important term using TRS
            if mmi <= nni:
                continue
            fnm = f_n[nni] - f_n[mmi]
            Emn = E_n[mmi] - E_n[nni] + fnm * eshift

            # Two band part
            if np.abs(fnm) > ftol:
                tmp = 2 * np.imag(
                    r_vnn[pol_v[0], nni, mmi]
                    * (rd_vvnn[pol_v[1], pol_v[2], mmi, nni]
                        + rd_vvnn[pol_v[2], pol_v[1], mmi, nni])) \
                    / (Emn * (2 * w_l - Emn))
                tmp += np.imag(
                    r_vnn[pol_v[1], mmi, nni]
                    * rd_vvnn[pol_v[2], pol_v[0], nni, mmi]
                    + r_vnn[pol_v[2], mmi, nni]
                    * rd_vvnn[pol_v[1], pol_v[0], nni, mmi]) \
                    / (Emn * (w_l - Emn))
                tmp += np.imag(
                    r_vnn[pol_v[0], nni, mmi]
                    * (r_vnn[pol_v[1], mmi, nni]
                        * D_vnn[pol_v[2], mmi, nni]
                        + r_vnn[pol_v[2], mmi, nni]
                        * D_vnn[pol_v[1], mmi, nni])) \
                    * (1 / (w_l - Emn)
                        - 4 / (2 * w_l - Emn)) / Emn**2
                tmp -= np.imag(
                    r_vnn[pol_v[1], mmi, nni]
                    * rd_vvnn[pol_v[0], pol_v[2], nni, mmi]
                    + r_vnn[pol_v[2], mmi, nni]
                    * rd_vvnn[pol_v[0], pol_v[1], nni, mmi]) \
                    / (2 * Emn * (w_l - Emn))
                sum2_l += 1j * fnm * tmp / 2  # 1j imag

            # Three band term
            for lli in band_n:
                fnl = f_n[nni] - f_n[lli]
                fml = f_n[mmi] - f_n[lli]
                Eml = E_n[mmi] - E_n[lli] - fml * eshift
                Eln = E_n[lli] - E_n[nni] + fnl * eshift
                # Do not do zero calculations
                if (np.abs(fnm) < ftol and np.abs(fnl) < ftol
                        and np.abs(fml) < ftol):
                    continue
                if np.abs(Eln - Eml) < Etol:
                    continue

                rnml = np.real(
                    r_vnn[pol_v[0], nni, mmi]
                    * (r_vnn[pol_v[1], mmi, lli]
                        * r_vnn[pol_v[2], lli, nni]
                        + r_vnn[pol_v[2], mmi, lli]
                        * r_vnn[pol_v[1], lli, nni])) / (2 * (Eln - Eml))
                if np.abs(fnm) > ftol:
                    sum3_l += 2 * fnm / (2 * w_l - Emn) * rnml
                if np.abs(fnl) > ftol:
                    sum3_l += -fnl / (w_l - Eln) * rnml
                if np.abs(fml) > ftol:
                    sum3_l += fml / (w_l - Eml) * rnml

    return sum2_l, sum3_l


def make_output(gauge, sum2_l, sum3_l):
    """
    Make the output in SI unit and return chi

    Input:
        gauge       Chosen gauge
        sum2_l      2-bands term
        sum3_l      3-bands term
    Output:
        chi_l       Output chi as an array
    """
    # Make the output in SI unit
    if gauge == 'lg':
        dim_ee = _e**3 / (_eps0 * (2.0 * np.pi)**3)
        dim_sum = (_hbar / (Bohr * 1e-10))**3 / \
            (_e**5 * (Bohr * 1e-10)**3) * (_hbar / _me)**3
        dim_SI = dim_ee * dim_sum
        chi_l = dim_SI * (1j * sum2_l + sum3_l)
    elif gauge == 'vg':
        # Make the output in SI unit
        dim_vg = _e**3 * _hbar**2 / (_me**3 * (2.0 * np.pi)**3)
        dim_chi = 1j * _hbar / (_eps0 * 2.0 * _e)  # 2 beacuse of frequecny
        dim_sum = (_hbar / (Bohr * 1e-10))**3 / \
            (_e**4 * (Bohr * 1e-10)**3)
        dim_SI = dim_chi * dim_vg * dim_sum
        chi_l = dim_SI * (sum2_l + sum3_l)
    else:
        parprint('Gauge ' + gauge + ' not implemented.')
        raise NotImplementedError

    return chi_l



import numpy as np

from ase.units import Hartree, Bohr

from gpaw.xc.hybrid import HybridXCBase


def calculate_forces(wfs, dens, ham, log=None):
    """Return the atomic forces."""

    assert not isinstance(ham.xc, HybridXCBase)
    assert not ham.xc.name.startswith('GLLB')

    natoms = len(wfs.setups)

    # Force from projector functions (and basis set):
    F_wfs_av = np.zeros((natoms, 3))
    wfs.calculate_forces(ham, F_wfs_av)
    wfs.gd.comm.sum(F_wfs_av, 0)
    F_ham_av = np.zeros((natoms, 3))

    try:
        # ODD functionals need force corrections for each spin
        correction = ham.xc.setup_force_corrections
    except AttributeError:
        pass
    else:
        correction(F_ham_av)

    ham.calculate_forces(dens, F_ham_av)

    F_av = F_ham_av + F_wfs_av
    wfs.world.broadcast(F_av, 0)

    F_av = wfs.kd.symmetry.symmetrize_forces(F_av)

    if log:
        log('\nForces in eV/Ang:')
        c = Hartree / Bohr
        for a, setup in enumerate(wfs.setups):
            log('%3d %-2s %10.5f %10.5f %10.5f' %
                ((a, setup.symbol) + tuple(F_av[a] * c)))
        log()

    return F_av



from collections.abc import MutableMapping
import numpy as np


class DefaultKeyMapping:
    def a2key(self, a):
        return a

    def key2a(self, key):
        return key


class TableKeyMapping:
    def __init__(self, a2key_a, key2a_k):
        self.a2key_a = a2key_a
        self.key2a_k = key2a_k

    def key2a(self, key):
        return self.key2a_k[key]

    def a2key(self, a):
        return self.a2key_a[a]


class ArrayDict(MutableMapping):
    """Distributed dictionary of fixed-size, fixed-dtype arrays.

    Implements a map [0, ..., N] -> [A0, ..., AN]

    Elements are initialized as empty numpy arrays.

    Unlike a normal dictionary, this class implements a strict loop ordering
    which is consistent with that of the underlying atom partition."""
    def __init__(self, partition, shapes_a, dtype=float, d=None,
                 keymap=None):
        self.partition = partition
        if callable(shapes_a):
            shapes_a = [shapes_a(a) for a in range(self.partition.natoms)]
        self.shapes_a = shapes_a  # global
        assert len(shapes_a) == partition.natoms
        self.dtype = dtype
        self.data = {}

        # This will be a terrible hack to make it easier to customize
        # the "keys".  Normally keys are 0...N and correspond to
        # rank_a of the AtomPartition.  But the keymap allows the user
        # to pass around a one-to-one mapping between the "physical"
        # keys 0...N and some other objects that the user may fancy.
        if keymap is None:
            keymap = DefaultKeyMapping()
        self.keymap = keymap

        if d is None:
            for a in partition.my_indices:
                self[a] = np.empty(self.shapes_a[a], dtype=dtype)
        else:
            self.update(d)
        self.check_consistency()

    # copy() is dangerous since redistributions write back
    # into arrays, and redistribution of a copy could lead to bugs
    # if the other copy changes.
    #
    # Let's just use deepcopy to stay out of trouble.  Then the
    # dictionary copy() is not invoked confusingly.
    #
    #  -askhl
    def copy(self):
        copy = ArrayDict(self.partition, self.shapes_a, dtype=self.dtype,
                         keymap=self.keymap)
        for a in self:
            copy[a][:] = self[a]
        return copy

    def __getitem__(self, a):
        value = self.data[a]
        assert value.shape == self.shapes_a[a]
        assert value.dtype == self.dtype
        return value

    def __setitem__(self, a, value):
        assert value.shape == self.shapes_a[a], \
            'defined shape %s vs new %s' % (self.shapes_a[a], value.shape)
        assert value.dtype == self.dtype
        self.data[a] = value

    def redistribute(self, partition):
        """Redistribute according to specified partition."""
        def get_empty(a):
            return np.empty(self.shapes_a[a], self.dtype)

        self.partition.redistribute(partition, self, get_empty)
        self.partition = partition  # Better with immutable partition?
        self.check_consistency()

    def check_consistency(self):
        k1 = set(self.partition.my_indices)
        k2 = set(self.data.keys())
        assert k1 == k2, 'Required keys %s different from actual %s' % (k1, k2)
        for a, array in self.items():
            assert array.dtype == self.dtype
            assert array.shape == self.shapes_a[a], \
                'array shape %s vs specified shape %s' % (array.shape,
                                                          self.shapes_a[a])

    def toarray(self, axis=None):
        # We could also implement it as a contiguous buffer.
        if len(self) == 0:
            # XXXXXX how should we deal with globally or locally empty arrays?
            # This will probably lead to bugs unless we get all the
            # dimensions right.
            return np.empty(0, self.dtype)
        if axis is None:
            return np.concatenate([self[a].ravel()
                                   for a in self.partition.my_indices])
        else:
            # XXX self[a].shape must all be consistent except along axis
            return np.concatenate([self[a] for a in self.partition.my_indices],
                                  axis=axis)

    def fromarray(self, data):
        assert data.dtype == self.dtype
        M1 = 0
        for a in self.partition.my_indices:
            M2 = M1 + np.prod(self.shapes_a[a])
            self[a].ravel()[:] = data[M1:M2]
            M1 = M2

    def redistribute_and_broadcast(self, dist_comm, dup_comm):
        # Data exists on self which is a "nice" distribution but now
        # we want it on sub_partition which has a smaller communicator
        # whose parent is self.comm.
        #
        # We want our own data replicated on each

        # XXX direct comparison of communicators are unsafe as we do not use
        # MPI_Comm_compare

        # assert subpartition.comm.parent == self.partition.comm
        from gpaw.utilities.partition import AtomPartition

        newrank_a = self.partition.rank_a % dist_comm.size
        masters_only_partition = AtomPartition(self.partition.comm, newrank_a)
        dst_partition = AtomPartition(dist_comm, newrank_a)
        copy = self.copy()
        copy.redistribute(masters_only_partition)

        dst = ArrayDict(dst_partition, self.shapes_a, dtype=self.dtype,
                        keymap=self.keymap)
        data = dst.toarray()
        if dup_comm.rank == 0:
            data0 = copy.toarray()
            data[:] = data0
        dup_comm.broadcast(data, 0)
        dst.fromarray(data)
        return dst

    def __iter__(self):
        # These functions enforce the same ordering as self.partition
        # when looping.
        return iter(self.partition.my_indices)

    def __repr__(self):
        tokens = []
        for key in sorted(self.keys()):
            shapestr = 'x'.join(map(str, self.shapes_a[key]))
            tokens.append('%s:%s' % (self.keymap.a2key(key), shapestr))
        text = ', '.join(tokens)
        return '%s@rank%d/%d {%s}' % (self.__class__.__name__,
                                      self.partition.comm.rank,
                                      self.partition.comm.size,
                                      text)

    def __delitem__(self, a):
        # Actually this is not quite right; we effectively delete items
        # when we redistribute the arraydict.  But this is another
        # code path so let's resolve this later if necessary.
        raise TypeError('Deleting arraydict elements not supported since '
                        'doing so violates the input list-of-shapes')

    def __len__(self):
        return len(self.data)



from pathlib import Path
from typing import Union

import numpy as np
from ase.units import Ha

from gpaw import GPAW
from gpaw.calculator import GPAW as Calculator
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.mpi import serial_comm
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.lfc import PWLFC
from gpaw.xc import XC
from . import parse_name
from .coulomb import coulomb_interaction
from .kpts import get_kpt
from .paw import calculate_paw_stuff
from .symmetry import Symmetry
from gpaw.typing import Array1D


def non_self_consistent_energy(calc: Union[Calculator, str, Path],
                               xcname: str,
                               ftol=1e-9) -> Array1D:
    """Calculate non self-consistent energy for Hybrid functional.

    Based on a self-consistent DFT calculation (calc).  EXX integrals involving
    states with occupation numbers less than ftol are skipped.

    >>> energies = non_self_consistent_energy('<gpw-file>',
    ...                                       xcname='HSE06')
    >>> e_hyb = energies.sum()

    The correction to the self-consistent energy will be
    ``energies[1:].sum()``.

    The returned energy contributions are (in eV):

    1. DFT total free energy (not extrapolated to zero smearing)
    2. minus DFT XC energy
    3. Hybrid semi-local XC energy
    4. EXX core-core energy
    5. EXX core-valence energy
    6. EXX valence-valence energy
    """

    if calc == '<gpw-file>':  # for doctest
        return np.zeros(6)

    if isinstance(calc, (str, Path)):
        calc = GPAW(calc, txt=None, parallel={'band': 1, 'kpt': 1})

    assert not isinstance(calc, (str, Path))  # for mypy
    wfs = calc.wfs
    dens = calc.density
    kd = wfs.kd
    setups = wfs.setups
    nspins = wfs.nspins

    nocc = max(((kpt.f_n / kpt.weight) > ftol).sum()
               for kpt in wfs.kpt_u)
    nocc = kd.comm.max(wfs.bd.comm.sum(int(nocc)))

    xcname, exx_fraction, omega = parse_name(xcname)

    xc = XC(xcname)
    exc = 0.0
    for a, D_sp in dens.D_asp.items():
        exc += xc.calculate_paw_correction(setups[a], D_sp)
    exc = dens.finegd.comm.sum(exc)
    if dens.nt_sg is None:
        dens.interpolate_pseudo_density()
    exc += xc.calculate(dens.finegd, dens.nt_sg)

    coulomb = coulomb_interaction(omega, wfs.gd, kd)
    sym = Symmetry(kd)

    paw_s = calculate_paw_stuff(wfs, dens)

    ecc = sum(setup.ExxC for setup in setups) * exx_fraction
    evc = 0.0
    evv = 0.0
    for spin in range(nspins):
        kpts = [get_kpt(wfs, k, spin, 0, nocc) for k in range(kd.nibzkpts)]
        e1, e2 = calculate_energy(kpts, paw_s[spin],
                                  wfs, sym, coulomb, calc.spos_ac)
        evc += e1 * exx_fraction * 2 / wfs.nspins
        evv += e2 * exx_fraction * 2 / wfs.nspins

    return np.array([calc.hamiltonian.e_total_free,
                     -calc.hamiltonian.e_xc,
                     exc,
                     ecc,
                     evc,
                     evv]) * Ha


def calculate_energy(kpts, paw, wfs, sym, coulomb, spos_ac):
    pd = kpts[0].psit.pd
    gd = pd.gd.new_descriptor(comm=serial_comm)
    comm = wfs.world

    exxvv = 0.0
    exxvc = 0.0
    for i, kpt in enumerate(kpts):
        for a, VV_ii in paw.VV_aii.items():
            P_ni = kpt.proj[a]
            vv_n = np.einsum('ni, ij, nj -> n',
                             P_ni.conj(), VV_ii, P_ni).real
            vc_n = np.einsum('ni, ij, nj -> n',
                             P_ni.conj(), paw.VC_aii[a], P_ni).real
            exxvv -= vv_n.dot(kpt.f_n) * kpt.weight
            exxvc -= vc_n.dot(kpt.f_n) * kpt.weight

    for i1, i2, s, k1, k2, count in sym.pairs(kpts, wfs, spos_ac):
        q_c = k2.k_c - k1.k_c
        qd = KPointDescriptor([-q_c])

        pd12 = PWDescriptor(pd.ecut, gd, pd.dtype, kd=qd)
        ghat = PWLFC([data.ghat_l for data in wfs.setups], pd12)
        ghat.set_positions(spos_ac)

        v_G = coulomb.get_potential(pd12)
        e_nn = calculate_exx_for_pair(k1, k2, ghat, v_G, comm,
                                      paw.Delta_aiiL)
        e_nn *= count
        e = k1.f_n.dot(e_nn).dot(k2.f_n) / sym.kd.nbzkpts**2
        exxvv -= 0.5 * e

    exxvv = comm.sum(exxvv)
    exxvc = comm.sum(exxvc)

    return exxvc, exxvv


def calculate_exx_for_pair(k1,
                           k2,
                           ghat,
                           v_G,
                           comm,
                           Delta_aiiL):

    N1 = len(k1.u_nR)
    N2 = len(k2.u_nR)

    size = comm.size
    rank = comm.rank

    Q_annL = [np.einsum('mi, ijL, nj -> mnL',
                        k1.proj[a],
                        Delta_iiL,
                        k2.proj[a].conj())
              for a, Delta_iiL in enumerate(Delta_aiiL)]

    if k1 is k2:
        n2max = (N1 + size - 1) // size
    else:
        n2max = N2

    e_nn = np.zeros((N1, N2))
    rho_nG = ghat.pd.empty(n2max, k1.u_nR.dtype)

    for n1, u1_R in enumerate(k1.u_nR):
        if k1 is k2:
            B = (N1 - n1 + size - 1) // size
            n2a = min(n1 + rank * B, N2)
            n2b = min(n2a + B, N2)
        else:
            n2a = 0
            n2b = N2

        for n2, rho_G in enumerate(rho_nG[:n2b - n2a], n2a):
            rho_G[:] = ghat.pd.fft(u1_R * k2.u_nR[n2].conj())

        ghat.add(rho_nG[:n2b - n2a],
                 {a: Q_nnL[n1, n2a:n2b]
                  for a, Q_nnL in enumerate(Q_annL)})

        for n2, rho_G in enumerate(rho_nG[:n2b - n2a], n2a):
            vrho_G = v_G * rho_G
            e = ghat.pd.integrate(rho_G, vrho_G).real
            e_nn[n1, n2] = e
            if k1 is k2:
                e_nn[n2, n1] = e

    return e_nn



import numpy as np


class Q0Correction:
    def __init__(self, cell_cv, bzk_kc, N_c):
        self.cell_cv = cell_cv
        self.bzk_kc = bzk_kc
        self.N_c = N_c

        # Check that basic assumptions of cell and k-grid
        # for Q0Correction are fulfilled
        assert N_c[2] == 1  # z-axis is non periodic direction
        eps = 1e-14
        assert (abs(cell_cv[:2, 2]).max() < eps and
                abs(cell_cv[2, :2]).max() < eps and
                cell_cv[2, 2] > 0)

        # Hardcoded crap?
        x0density = 0.1  # ? 0.01

        # Generate numerical q-point grid
        self.rcell_cv = 2 * np.pi * np.linalg.inv(cell_cv).T

        # Prepare stuff
        self.q0cell_cv = np.array([1, 1, 1])**0.5 * self.rcell_cv / self.N_c
        L = cell_cv[2, 2]
        q0density = 2. / L * x0density
        npts_c = np.ceil(np.sum(self.q0cell_cv**2, axis=1)**0.5 /
                         q0density).astype(int)
        npts_c[2] = 1
        npts_c += (npts_c + 1) % 2
        self.npts_c = npts_c

    def add_q0_correction(self, qpd, W_GG, einv_GG,
                          chi0_xvG, chi0_vv, sqrtV_G):
        from ase.dft import monkhorst_pack
        qpts_qc = self.bzk_kc
        pi = np.pi
        L = self.cell_cv[2, 2]

        vc_G0 = sqrtV_G[1:]**2

        B_GG = einv_GG[1:, 1:]
        u_v0G = vc_G0[np.newaxis, :]**0.5 * chi0_xvG[0, :, 1:]
        u_vG0 = vc_G0[np.newaxis, :]**0.5 * chi0_xvG[1, :, 1:]
        U_vv = -chi0_vv
        a_v0G = -np.dot(u_v0G, B_GG)
        a_vG0 = -np.dot(u_vG0, B_GG.T)
        A_vv = U_vv - np.dot(a_v0G, u_vG0.T)
        S_v0G = a_v0G
        S_vG0 = a_vG0
        L_vv = A_vv

        # Get necessary G vectors.
        G_Gv = qpd.get_reciprocal_vectors(add_q=False)[1:]
        G_Gv += np.array([1e-14, 1e-14, 0])
        G2_G = np.sum(G_Gv**2, axis=1)
        Gpar_G = np.sum(G_Gv[:, 0:2]**2, axis=1)**0.5

        # There is still a lot of stuff here,
        # which could go to the constructor! XXX
        iq = np.argmin(np.sum(qpts_qc**2, axis=1))
        assert np.allclose(qpts_qc[iq], 0)
        q0vol = abs(np.linalg.det(self.q0cell_cv))

        qpts_qc = monkhorst_pack(self.npts_c)
        qgamma = np.argmin(np.sum(qpts_qc**2, axis=1))

        qpts_qv = np.dot(qpts_qc, self.q0cell_cv)
        qpts_q = np.sum(qpts_qv**2, axis=1)**0.5
        qpts_q[qgamma] = 1e-14
        qdir_qv = qpts_qv / qpts_q[:, np.newaxis]
        qdir_qvv = qdir_qv[:, :, np.newaxis] * qdir_qv[:, np.newaxis, :]
        nq = len(qpts_qc)
        q0area = q0vol / self.q0cell_cv[2, 2]
        dq0 = q0area / nq
        dq0rad = (dq0 / pi)**0.5
        R = L / 2.
        x0area = q0area * R**2
        dx0rad = dq0rad * R

        exp_q = 4 * pi * (1 - np.exp(-qpts_q * R))
        dv_G = ((pi * L * G2_G * np.exp(-Gpar_G * R) * np.cos(G_Gv[:, 2] * R) -
                 4 * pi * Gpar_G * (1 - np.exp(-Gpar_G * R) *
                                    np.cos(G_Gv[:, 2] * R))) /
                (G2_G**1.5 * Gpar_G *
                 (4 * pi * (1 - np.exp(-Gpar_G * R) *
                            np.cos(G_Gv[:, 2] * R)))**0.5))

        dv_Gv = dv_G[:, np.newaxis] * G_Gv

        # Add corrections
        W_GG[:, 0] = 0.0
        W_GG[0, :] = 0.0

        A_q = np.sum(qdir_qv * np.dot(qdir_qv, L_vv), axis=1)
        frac_q = 1. / (1 + exp_q * A_q)

        # HEAD:
        w00_q = -(exp_q / qpts_q)**2 * A_q * frac_q
        w00_q[qgamma] = 0.0
        W_GG[0, 0] = w00_q.sum() / nq
        Axy = 0.5 * (L_vv[0, 0] + L_vv[1, 1])  # in-plane average
        a0 = 4 * pi * Axy + 1

        W_GG[0, 0] += -((a0 * dx0rad - np.log(a0 * dx0rad + 1)) /
                        a0**2 / x0area * 2 * np.pi * (2 * pi * L)**2 * Axy)

        # WINGS:
        u_q = -exp_q / qpts_q * frac_q
        W_GG[1:, 0] = 1. / nq * np.dot(
            np.sum(qdir_qv * u_q[:, np.newaxis], axis=0),
            S_vG0 * sqrtV_G[np.newaxis, 1:])

        W_GG[0, 1:] = 1. / nq * np.dot(
            np.sum(qdir_qv * u_q[:, np.newaxis], axis=0),
            S_v0G * sqrtV_G[np.newaxis, 1:])

        # BODY:
        # Constant corrections:
        W_GG[1:, 1:] += 1. / nq * sqrtV_G[1:, None] * sqrtV_G[None, 1:] * \
            np.tensordot(S_v0G, np.dot(S_vG0.T,
                                       np.sum(-qdir_qvv *
                                              exp_q[:, None, None] *
                                              frac_q[:, None, None],
                                              axis=0)), axes=(0, 1))
        u_vvv = np.tensordot(u_q[:, None] * qpts_qv, qdir_qvv, axes=(0, 0))
        # Gradient corrections:
        W_GG[1:, 1:] += 1. / nq * np.sum(
            dv_Gv[:, :, None] * np.tensordot(
                S_v0G, np.tensordot(u_vvv, S_vG0 * sqrtV_G[None, 1:],
                                    axes=(2, 0)), axes=(0, 1)), axis=1)



import pytest
from io import StringIO

from ase.io import read
from ase.units import Ha

from gpaw import GPAW, FermiDirac
from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft2 import LrTDDFT2


jend = 12  # LUMO


@pytest.fixture
def C3H6O():
    atoms = read(StringIO("""10
http://cccbdb.nist.gov/ Geometry for C3H6O (Propylene oxide), CISD/6-31G*
O   0.8171  -0.7825  -0.2447
C  -1.5018   0.1019  -0.1473
H  -1.3989   0.3323  -1.2066
H  -2.0652  -0.8262  -0.0524
H  -2.0715   0.8983   0.3329
C  -0.1488  -0.0393   0.4879
H  -0.1505  -0.2633   1.5506
C   1.0387   0.6105  -0.0580
H   0.9518   1.2157  -0.9531
H   1.8684   0.8649   0.5908
"""), format='xyz')
    atoms.center(vacuum=3)

    atoms.calc = GPAW(mode='fd', h=0.3,
                      occupations=FermiDirac(width=0.1),
                      nbands=15, convergence={
                          'eigenstates': 1e-4,
                          'bands': jend},
                      txt=None)
    atoms.get_potential_energy()
    return atoms


@pytest.mark.later
def test_lrtddft2(C3H6O, in_tmp_dir):
    """Test equivalence"""
    atoms = C3H6O

    istart = 10  # HOMO-1

    evs = atoms.calc.get_eigenvalues() / Ha
    energy_differences = evs[jend] - evs[istart:jend]

    lr = LrTDDFT(atoms.calc, restrict={'istart': istart, 'jend': jend})

    lr2 = LrTDDFT2('C3H6O_lr', atoms.calc, fxc='LDA',
                   min_occ=istart, max_unocc=jend)
    lr2.calculate()

    # check for Kohn-Sham properties

    for de, ks, ks2 in zip(energy_differences,
                           lr.kss, lr2.ks_singles.kss_list[::-1]):
        assert de == pytest.approx(ks.energy, 1e-10)
        assert de == pytest.approx(ks2.energy_diff, 1e-8)

        assert ks.mur == pytest.approx(ks2.dip_mom_r, 1e-3)
        assert ks.magn == pytest.approx(ks2.magn_mom, 1e-6)

    # check for TDDFT properties

    (w, S, R, Sx, Sy, Sz) = lr2.get_transitions()

    assert len(lr) == len(w)
    for i, ex in enumerate(lr):
        assert ex.energy * Ha == pytest.approx(w[i], 1e-6)
        f = ex.get_oscillator_strength()
        assert f[0] == pytest.approx(S[i], 1e-4)
        assert f[1] == pytest.approx(Sx[i], 1e-3)
        assert f[2] == pytest.approx(Sy[i], 1e-3)
        assert f[3] == pytest.approx(Sz[i], 1e-3)
        assert ex.get_rotatory_strength() == pytest.approx(R[i], 1e-4)



import numpy as np
import pytest
from gpaw.core.matrix import Matrix
from gpaw.gpu import cupy as cp, as_xp


@pytest.mark.gpu
@pytest.mark.serial
def test_zyrk():
    a = np.array([[1, 1 + 2j, 2], [1, 0.5j, -1 - 0.5j]])
    m = Matrix(2, 3, data=a)
    b = m.multiply(m, opb='C', beta=0.0, symmetric=True)
    b.tril2full()
    a = cp.asarray(a)
    m = Matrix(2, 3, data=a)
    b2 = m.multiply(m, opb='C', beta=0.0, symmetric=True)
    b2.tril2full()
    c = b2.to_cpu()
    assert (c.data == b.data).all()


@pytest.mark.gpu
@pytest.mark.serial
def test_eigh():
    H1 = Matrix(2, 2, data=np.array([[2, 42.1 + 42.1j], [0.1 - 0.1j, 3]]))
    S1 = Matrix(2, 2, data=np.array([[1, 42.1 + 42.2j], [0.1 - 0.2j, 0.9]]))
    H2 = Matrix(2, 2, data=cp.asarray(H1.data))
    S2 = Matrix(2, 2, data=cp.asarray(S1.data))

    E1 = H1.eigh(S1)

    S0 = S1.copy()
    S0.tril2full()

    E2 = H2.eigh(S2)
    assert as_xp(E2, np) == pytest.approx(E1)

    C1 = H1.data
    C2 = H2.to_cpu().data

    # Check that eigenvectors are parallel:
    X = C1.conj() @ S0.data @ C2.T
    assert abs(X) == pytest.approx(np.eye(2))



"""Module defining  ``Eigensolver`` classes."""


import numpy as np
from gpaw.eigensolvers.eigensolver import Eigensolver
from gpaw.matrix_descriptor import MatrixDescriptor


class DirectPW(Eigensolver):
    """Direct eigensolver for plane waves"""

    def __init__(self, keep_htpsit=False):
        """Initialize direct eigensolver. """

        Eigensolver.__init__(self, keep_htpsit)

    def iterate_one_k_point(self, ham, wfs, kpt, weights):
        """Setup H and S matries and diagonalize for the kpoint"""

        self.timer.start('DirectPW')
        H_GG, S_GG = wfs.hs(ham, kpt.q, kpt.s)
        npw = len(H_GG)
        eps_n = np.empty(npw)
        md = MatrixDescriptor(npw, npw)
        psit_nG = md.empty(dtype=complex)
        md.general_diagonalize_dc(H_GG, S_GG, psit_nG, eps_n,
                                  iu=wfs.bd.nbands)
        kpt.eps_n[:] = eps_n[:wfs.bd.nbands]
        wfs.pt.integrate(kpt.psit_nG, kpt.P_ani, kpt.q)
        self.timer.stop('DirectPW')
        error = 0.0
        return error

    def __repr__(self):
        return 'DirectPW eigensolver'



import numpy as np
from ase.units import Ha
from gpaw.xc.functional import XCFunctional


class NonLocalFunctional(XCFunctional):
    def __init__(self, xcname, setup_name=None):
        self.contributions = []
        # TODO: remove self.xcs once deprecated calculate_delta_xc()
        # in c_response.py is removed
        self.xcs = {}
        XCFunctional.__init__(self, xcname, 'GLLB')
        if setup_name is None:
            self.setup_name = self.name
        else:
            self.setup_name = setup_name
        self.mix = None
        self.mix_vt_sg = None
        self.old_vt_sg = None
        self.old_H_asp = {}
        self.response = None

    def get_setup_name(self):
        return self.setup_name

    def set_mix(self, mix):
        self.mix = mix

    def initialize(self, density, hamiltonian, wfs):
        for contribution in self.contributions:
            contribution.initialize(density, hamiltonian, wfs)

    def initialize_1d(self, ae):
        for contribution in self.contributions:
            contribution.initialize_1d(ae)

    def calculate_impl(self, gd, n_sg, v_sg, e_g):
        e_g[:] = 0.0
        if self.mix is None:
            for contribution in self.contributions:
                contribution.calculate(e_g, n_sg, v_sg)
        else:
            cmix = self.mix
            if self.mix_vt_sg is None:
                self.mix_vt_sg = np.zeros_like(v_sg)
                self.old_vt_sg = np.zeros_like(v_sg)
                cmix = 1.0
            self.mix_vt_sg[:] = 0.0
            for contribution in self.contributions:
                contribution.calculate(e_g, n_sg, self.mix_vt_sg)
            self.mix_vt_sg = (cmix * self.mix_vt_sg
                              + (1.0 - cmix) * self.old_vt_sg)
            v_sg += self.mix_vt_sg
            self.old_vt_sg[:] = self.mix_vt_sg

    def calculate_paw_correction(self, setup, D_sp, dEdD_sp, a=None,
                                 addcoredensity=True):
        return self.calculate_energy_and_derivatives(setup, D_sp, dEdD_sp, a,
                                                     addcoredensity)

    def calculate_energy_and_derivatives(self, setup, D_sp, H_sp, a,
                                         addcoredensity=True):
        if setup.xc_correction is None:
            return 0.0
        Exc = 0.0
        # We are supposed to add to H_sp, not write directly
        H0_sp = H_sp
        H_sp = H0_sp.copy()
        H_sp[:] = 0.0

        if self.mix is None:
            for contribution in self.contributions:
                Exc += contribution.calculate_energy_and_derivatives(
                    setup, D_sp, H_sp, a, addcoredensity)
        else:
            cmix = self.mix
            if a not in self.old_H_asp:
                self.old_H_asp[a] = H_sp.copy()
                cmix = 1.0

            for contribution in self.contributions:
                Exc += contribution.calculate_energy_and_derivatives(
                    setup, D_sp, H_sp, a, addcoredensity)
            H_sp *= cmix
            H_sp += (1 - cmix) * self.old_H_asp[a]
            self.old_H_asp[a][:] = H_sp.copy()

        H0_sp += H_sp
        Exc -= setup.xc_correction.e_xc0
        return Exc

    def get_xc_potential_and_energy_1d(self, v_g):
        Exc = 0.0
        for contribution in self.contributions:
            Exc += contribution.add_xc_potential_and_energy_1d(v_g)
        return Exc

    def get_smooth_xc_potential_and_energy_1d(self, vt_g):
        Exc = 0.0
        for contribution in self.contributions:
            Exc += contribution.add_smooth_xc_potential_and_energy_1d(vt_g)
        return Exc

    def initialize_from_atomic_orbitals(self, basis_functions):
        for contribution in self.contributions:
            contribution.initialize_from_atomic_orbitals(basis_functions)

    def get_extra_setup_data(self, extra_data):
        for contribution in self.contributions:
            contribution.get_extra_setup_data(extra_data)

    def add_contribution(self, contribution):
        self.contributions.append(contribution)
        self.xcs[contribution.get_name()] = contribution
        if contribution.get_name() == 'RESPONSE':
            assert self.response is None
            self.response = contribution

    def get_description(self):
        fmt = '| %-6s | %-10s | %-45s |'
        header = fmt % ('Weight', 'Module', 'Description')
        dashes = '-' * len(header)
        s = ['{} functional being used consists of'.format(self.name)]
        s += [dashes]
        s += [header]
        s += [dashes]
        for c in self.contributions:
            s += [fmt % ('%6.3f' % c.weight, c.get_name(), c.get_desc())]
        s += [dashes]
        return '\n'.join(s)

    def summary(self, log):
        if self.response is not None:
            eref_s = self.response.eref_s
            source_s = self.response.eref_source_s
            if eref_s is not None and source_s is not None:
                text_s = [f'{eref * Ha:.5f} eV (from {source})'
                          for eref, source in zip(eref_s, source_s)]
                log('{} response reference energy: {}\n'
                    .format(self.name, ', '.join(text_s)))

    def read(self, reader):
        for contribution in self.contributions:
            contribution.read(reader)

    def write(self, writer):
        for contribution in self.contributions:
            contribution.write(writer)

    def heeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeelp(self, olddens):
        # XXX This function should be removed once the deprecated
        # `fixdensity=True` option is removed.
        for contribution in self.contributions:
            try:
                contribution.heeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeelp(olddens)
            except AttributeError:
                pass



from gpaw.tddft import TDDFT
from gpaw.tddft.ehrenfest import EhrenfestVelocityVerlet
from gpaw.test import equal
import pytest


@pytest.mark.later
def test_tddft_ehrenfest_nacl(in_tmp_dir, gpw_files):

    td_calc = TDDFT(gpw_files['nacl_fd'], propagator='EFSICN')
    evv = EhrenfestVelocityVerlet(td_calc, 0.001)

    i = 0
    evv.get_energy()
    r = evv.x[1][2] - evv.x[0][2]
    # print 'E = ', [i, r, evv.Etot, evv.Ekin, evv.e_coulomb]

    for i in range(5):
        evv.propagate(1.0)
        evv.get_energy()
        r = evv.x[1][2] - evv.x[0][2]
        print('E = ', [i + 1, r, evv.Etot, evv.Ekin, evv.e_coulomb])

    equal(r, 7.558883144, 1e-6)
    equal(evv.Etot, -0.10359175317017391, 1e-4)



from typing import TYPE_CHECKING

from gpaw.typing import Array1D, ArrayND
from gpaw.gpu import cupy as cp

__all__ = ['GPU_AWARE_MPI']

GPU_AWARE_MPI = False


def add_to_density(f: float,
                   psit_X: ArrayND,
                   nt_X: ArrayND) -> None:
    nt_X += f * abs(psit_X)**2


def pw_precond(G2_G: Array1D,
               r_G: Array1D,
               ekin: float,
               o_G: Array1D) -> None:
    x = 1 / ekin / 3 * G2_G
    a = 27.0 + x * (18.0 + x * (12.0 + x * 8.0))
    xx = x * x
    o_G[:] = -4.0 / 3 / ekin * a / (a + 16.0 * xx * xx) * r_G


def pw_insert(coef_G: Array1D,
              Q_G: Array1D,
              x: float,
              array_Q: Array1D) -> None:
    array_Q[:] = 0.0
    array_Q.ravel()[Q_G] = x * coef_G


def pw_insert_gpu(psit_nG,
                  Q_G,
                  scale,
                  psit_bQ):
    assert scale == 1.0
    psit_bQ[:, Q_G] = psit_nG


def pwlfc_expand(f_Gs, emiGR_Ga, Y_GL,
                 l_s, a_J, s_J,
                 cc, f_GI):
    """
    f_GI = xp.empty((G2 - G1, self.nI), complex)
    I1 = 0
    for J, (a, s) in enumerate(zip(self.a_J, self.s_J)):
        l = self.l_s[s]
        I2 = I1 + 2 * l + 1
        f_GI[:, I1:I2] = (f_Gs[:, s] *
                          emiGR_Ga[:, a] *
                          Y_GL[:, l**2:(l + 1)**2].T *
                          (-1.0j)**l).T
        I1 = I2
    if cc:
        f_GI = f_GI.conj()
    if self.dtype == float:
        f_GI = f_GI.T.copy().view(float).T.copy()

    return f_GI
    """
    raise NotImplementedError


def pwlfc_expand_gpu(f_Gs, emiGR_Ga, Y_GL,
                     l_s, a_J, s_J,
                     cc, f_GI, I_J):
    raise NotImplementedError


def dH_aii_times_P_ani_gpu(dH_aii, ni_a,
                           P_nI, out_nI):
    I1 = 0
    J1 = 0
    for ni in ni_a._data:
        I2 = I1 + ni
        J2 = J1 + ni**2
        dH_ii = dH_aii[J1:J2].reshape((ni, ni))
        out_nI[:, I1:I2] = P_nI[:, I1:I2] @ dH_ii
        I1 = I2
        J1 = J2


def add_to_density_gpu(weight_n, psit_nR, nt_R):
    for weight, psit_R in zip(weight_n, psit_nR):
        nt_R += float(weight) * cp.abs(psit_R)**2


def symmetrize_ft(a_R, b_R, r_cc, t_c, offset_c):
    raise NotImplementedError


if not TYPE_CHECKING:
    try:
        from _gpaw import (  # noqa
            add_to_density, pw_precond, pw_insert,
            pwlfc_expand, symmetrize_ft)
    except ImportError:
        pass
    try:
        from _gpaw import gpu_aware_mpi as GPU_AWARE_MPI
    except ImportError:
        pass
    try:
        from _gpaw import (  # noqa
            pwlfc_expand_gpu, add_to_density_gpu, pw_insert_gpu,
            dH_aii_times_P_ani_gpu)
    except ImportError:
        pass



import numpy as np
import numbers
from scipy.optimize import minimize
from scipy.integrate import simps
from gpaw import GPAW, PW
from ase.parallel import parprint
import scipy.integrate as integrate
from scipy.interpolate import InterpolatedUnivariateSpline
from ase.units import Hartree as Ha
from ase.units import Bohr


class ElectrostaticCorrections():
    """
    Calculate the electrostatic corrections for charged defects.
    """
    def __init__(self, pristine, charged,
                 q=None, sigma=None, r0=None, dimensionality='3d'):
        if isinstance(pristine, str):
            pristine = GPAW(pristine, txt=None, parallel={'domain': 1})
        if isinstance(charged, str):
            charged = GPAW(charged, txt=None)
        calc = GPAW(mode=PW(500, force_complex_dtype=True),
                    kpts={'size': (1, 1, 1),
                          'gamma': True},
                    parallel={'domain': 1},
                    symmetry='off',
                    txt=None)
        atoms = pristine.atoms.copy()
        calc.initialize(atoms)

        self.pristine = pristine
        self.charged = charged
        self.calc = calc
        self.q = q
        self.sigma = sigma
        self.dimensionality = dimensionality

        self.pd = self.calc.wfs.pd
        self.L = self.pd.gd.cell_cv[2, 2]
        if r0 is not None:
            self.r0 = np.array(r0) / Bohr
        elif dimensionality == '2d':
            self.r0 = np.array([0, 0, self.L / 2])
        else:
            self.r0 = np.array([0, 0, 0])
        self.z0 = self.r0[2]
        self.G_Gv = self.pd.get_reciprocal_vectors(q=0, add_q=False)
        self.G2_G = self.pd.G2_qG[0]  # |\vec{G}|^2 in Bohr^-2
        self.rho_G = self.calculate_gaussian_density()
        self.Omega = abs(np.linalg.det(self.calc.density.gd.cell_cv))
        self.data = None
        self.El = None

        # For the 2D case, we assume that the dielectric profile epsilon(z)
        # follows the electronic density of the pristine system n(z).
        density = self.pristine.get_pseudo_density(gridrefinement=2)
        density_1d = density.mean(axis=(0, 1))
        coarse_z = find_z(self.pristine.density.gd.refine())
        fine_z = find_z(self.pd.gd.refine())
        transformer = InterpolatedUnivariateSpline(coarse_z, density_1d)
        self.density_1d = np.array([transformer(x) for x in fine_z])

        # We potentially have two z axes -- one on the original calculation,
        # and one on the calculator we just set up.
        self.z_g = fine_z
        self.density_z = coarse_z

        self.G_z = find_G_z(self.G_Gv)
        self.G_parallel = np.unique(self.G_Gv[:, :2], axis=0)
        self.GG = np.outer(self.G_z, self.G_z)  # G * Gprime

        # We need the G_z vectors on the finely sampled grid, to calculate
        # epsilon(G - G'), which goes up to 2G_max in magnitude.
        self.G_z_fine = (np.fft.fftfreq(len(fine_z)) *
                         2 * np.pi / (self.L) * len(fine_z))
        self.index_array = self.get_index_array()

        self.epsilons_z = None
        self.Eli = None
        self.Elp = None
        self.epsilon_GG = None

    def calculate_gaussian_density(self):
        # Fourier transformed gaussian:
        return self.q * np.exp(-0.5 * self.G2_G * self.sigma ** 2)

    def set_epsilons(self, epsilons, epsilon_bulk=1):
        """Set the bulk dielectric constant of the system corresponding to the
        atomic configuration of the pristine system. This is used to calculate
        the screened coulomb potential of the gaussian model distribution that
        we use. In 2D, the dielectric constant is poorly defined. However, when
        we do DFT calculations with periodic systems, we are always working
        with pseudo bulk-like structures, due to the periodic boundary
        conditions. It is therefore possible to calculate some dielectric
        constant. The screening of the system must come from where there is an
        electronic density, and we thus set

        epsilon(z) = k * n(z) + epsilon_bulk,

        for some constant of proprotionality k. This is determined by the
        constraint that the total screening int dz epsilon(z) gives the correct
        value.

        For 3d systems, the trick is to set k = 0. In that case, we have the
        constant dielectric function that we require.

        Parameters:
          epsilons: A float, or a list of two floats. If a single
            float is given, the screening in-plane and out-of-plane is assumed
            to be the same. If two floats are given, the first represents the
            in-plane response, and the second represents the out-of-plane
            response.
          epsilon_bulk: The value of the screening far away from the system,
            given in the same format as epsilons. In most cases, this will be
            1, corresponding to vacuum.

        Returns:
            epsilons_z: dictionary containnig the normalized dielectric
              profiles in the in-plane and out-of-plane directions along the z
              axis.
        """

        # Reset the state of the object; the corrections need to be
        # recalculated
        self.Elp = None
        self.Eli = None
        self.epsilon_GG = None

        def normalize(value):
            if isinstance(value, numbers.Number):
                value = [value, value]
            elif len(value) == 1:
                value = [value[0]] * 2
            return np.array(value)

        epsilons = normalize(epsilons)
        self.epsilons = epsilons

        if self.dimensionality == '2d':
            eb = normalize(epsilon_bulk)
        elif self.dimensionality == '3d':
            eb = normalize(epsilons)
        self.eb = eb

        z = self.z_g
        L = self.L
        density_1d = self.density_1d
        N = simps(density_1d, z)
        epsilons_z = np.zeros((2,) + np.shape(density_1d))
        epsilons_z += density_1d

        # In-plane
        epsilons_z[0] = epsilons_z[0] * (epsilons[0] - eb[0]) / N * L + eb[0]

        # Out-of-plane
        def objective_function(k):
            k = k[0]
            integral = simps(1 / (k * density_1d + eb[1]), z) / L
            return np.abs(integral - 1 / epsilons[1])

        test = minimize(objective_function, [1], method='Nelder-Mead')
        assert test.success, "Unable to normalize dielectric profile"
        k = test.x[0]
        epsilons_z[1] = epsilons_z[1] * k + eb[1]
        self.epsilons_z = {'in-plane': epsilons_z[0],
                           'out-of-plane': epsilons_z[1]}
        self.calculate_epsilon_GG()

    def get_index_array(self):
        """
        Calculate the indices that map between the G vectors on the fine grid,
        and the matrix defined by M(G, Gprime) = G - Gprime.

        We use this to generate the matrix epsilon_GG = epsilon(G - Gprime)
        based on knowledge of epsilon(G) on the fine grid.
        """
        G, Gprime = np.meshgrid(self.G_z, self.G_z)
        difference_GG = G - Gprime
        index_array = np.zeros(np.shape(difference_GG))
        index_array[:] = np.nan
        for idx, val in enumerate(self.G_z_fine):
            mask = np.isclose(difference_GG, val)
            index_array[mask] = idx
        if np.isnan(index_array).any():
            print("Missing indices found in mapping between plane wave "
                  "sets. Something is wrong with your G-vectors!")
            print(np.isnan(index_array).all())
            print(np.where(np.isnan(index_array)))
            print(self.G_z_fine)
            print(difference_GG)
            assert False
        return index_array.astype(int)

    def calculate_epsilon_GG(self):
        assert self.epsilons_z is not None, ("You provide the dielectric "
                                             "constant first!")
        N = len(self.density_1d)
        epsilon_par_G = np.fft.fft(self.epsilons_z['in-plane']) / N
        epsilon_perp_G = np.fft.fft(self.epsilons_z['out-of-plane']) / N
        epsilon_par_GG = epsilon_par_G[self.index_array]
        epsilon_perp_GG = epsilon_perp_G[self.index_array]

        self.epsilon_GG = {'in-plane': epsilon_par_GG,
                           'out-of-plane': epsilon_perp_GG}

    def calculate_periodic_correction(self):
        G_z = self.G_z
        if self.Elp is not None:
            return self.Elp
        if self.epsilon_GG is None:
            self.calculate_epsilon_GG()
        Elp = 0.0

        for vector in self.G_parallel:
            G2 = (np.dot(vector, vector) + G_z * G_z)
            rho_G = self.q * np.exp(- G2 * self.sigma ** 2 / 2, dtype=complex)
            if self.dimensionality == '2d':
                rho_G *= np.exp(1j * self.z0 * G_z)
            A_GG = (self.GG * self.epsilon_GG['out-of-plane']
                    + np.dot(vector, vector) * self.epsilon_GG['in-plane'])
            if np.allclose(vector, 0):
                A_GG[0, 0] = 1  # The d.c. potential is poorly defined
            V_G = np.linalg.solve(A_GG, rho_G)
            if np.allclose(vector, 0):
                parprint('Skipping G^2=0 contribution to Elp')
                V_G[0] = 0
            Elp += (rho_G * V_G).sum().real

        Elp *= 2.0 * np.pi * Ha / self.Omega

        self.Elp = Elp
        return Elp

    def calculate_isolated_correction(self):
        if self.Eli is not None:
            return self.Eli

        L = self.L
        G_z = self.G_z

        G, Gprime = np.meshgrid(G_z, G_z)
        Delta_GG = G - Gprime
        phase = Delta_GG * self.z0
        gaussian = (Gprime ** 2 + G ** 2) * self.sigma * self.sigma / 2

        prefactor = np.exp(1j * phase - gaussian)

        if self.epsilon_GG is None:
            self.calculate_epsilon_GG()

        dE_GG_par = (self.epsilon_GG['in-plane']
                     - self.eb[0] * np.eye(len(self.G_z)))
        dE_GG_perp = (self.epsilon_GG['out-of-plane']
                      - self.eb[1] * np.eye(len(self.G_z)))

        def integrand(k):
            K_inv_G = ((self.eb[0] * k ** 2 + self.eb[1] * G_z ** 2) /
                       (1 - np.exp(-k * L / 2) * np.cos(L * G_z / 2)))
            K_inv_GG = np.diag(K_inv_G)
            D_GG = L * (K_inv_GG + dE_GG_perp * self.GG + dE_GG_par * k ** 2)
            return (k * np.exp(-k ** 2 * self.sigma ** 2) *
                    (prefactor * np.linalg.inv(D_GG)).sum()).real

        I = integrate.quad(integrand, 0, np.inf, limit=500)
        Eli = self.q * self.q * I[0] * Ha
        self.Eli = Eli
        return Eli

    def calculate_potential_alignment(self):
        V_neutral = -np.mean(self.pristine.get_electrostatic_potential(),
                             (0, 1))
        V_charged = -np.mean(self.charged.get_electrostatic_potential(),
                             (0, 1))

        z = self.density_z
        z_model, V_model = self.calculate_z_avg_model_potential()
        V_model = InterpolatedUnivariateSpline(z_model[:-1], V_model[:-1])
        V_model = V_model(z)
        self.V_model = V_model
        Delta_V = self.average(V_model - V_charged + V_neutral, z)
        return Delta_V

    def calculate_z_avg_model_potential(self):

        vox3 = self.calc.density.gd.cell_cv[2, :] / len(self.z_g)

        # The grid is arranged with z increasing fastest, then y
        # then x (like a cube file)

        G_z = self.G_z[1:]
        rho_Gz = self.q * np.exp(-0.5 * G_z * G_z * self.sigma * self.sigma)

        zs = []
        Vs = []
        if self.dimensionality == '2d':
            phase = np.exp(1j * (self.G_z * self.z0))
            A_GG = (self.GG * self.epsilon_GG['out-of-plane'])
            A_GG[0, 0] = 1
            V_G = np.linalg.solve(A_GG,
                                  phase * np.array([0] + list(rho_Gz)))[1:]
        elif self.dimensionality == '3d':
            phase = np.exp(1j * (G_z * self.z0))
            V_G = phase * rho_Gz / self.eb[1] / G_z ** 2

        for z in self.z_g:
            phase_G = np.exp(1j * (G_z * z))
            V = (np.sum(phase_G * V_G).real
                 * Ha * 4.0 * np.pi / (self.Omega))
            Vs.append(V)

        V = (np.sum(V_G.real) * Ha * 4.0 * np.pi / (self.Omega))
        zs = list(self.z_g) + [vox3[2]]
        Vs.append(V)
        return np.array(zs), np.array(Vs)

    def average(self, V, z):
        N = len(V)
        if self.dimensionality == '3d':
            middle = np.argmin(np.abs(z - self.z0)) + N // 2
            middle = middle % len(z)
            points = range(middle - N // 8, middle + N // 8 + 1)
            restricted = V[points]
        elif self.dimensionality == '2d':
            points = list(range(0, N // 8)) + list(range(7 * N // 8, N))
        restricted = V[points]
        V_mean = np.mean(restricted)
        return V_mean

    def calculate_corrected_formation_energy(self):
        E_0 = self.pristine.get_potential_energy()
        E_X = self.charged.get_potential_energy()
        Eli = self.calculate_isolated_correction()
        Elp = self.calculate_periodic_correction()
        Delta_V = self.calculate_potential_alignment()
        return E_X - E_0 - (Elp - Eli) + Delta_V * self.q

    def calculate_uncorrected_formation_energy(self):
        E_0 = self.pristine.get_potential_energy()
        E_X = self.charged.get_potential_energy()
        return E_X - E_0

    def collect_electrostatic_data(self):
        V_neutral = -np.mean(self.pristine.get_electrostatic_potential(),
                             (0, 1)),
        V_charged = -np.mean(self.charged.get_electrostatic_potential(),
                             (0, 1)),
        data = {'epsilon': self.eb[0],
                'z': self.density_z,
                'V_0': V_neutral,
                'V_X': V_charged,
                'Elc': self.Elp - self.Eli,
                'D_V_mean': self.calculate_potential_alignment(),
                'V_model': self.V_model,
                'D_V': (self.V_model
                        + V_neutral
                        - V_charged)}
        self.data = data
        return data


def find_G_z(G_Gv):
    mask = (G_Gv[:, 0] == 0) & (G_Gv[:, 1] == 0)
    G_z = G_Gv[mask][:, 2]  # qG_z vectors in Bohr^{-1}
    return G_z


def find_z(gd):
    r3_xyz = gd.get_grid_point_coordinates()
    nrz = r3_xyz.shape[3]
    return r3_xyz[2].flatten()[:nrz]



"""Check for change in total energy and lowest eigenvalue regarding to box."""
from ase import Atoms
from ase.parallel import parprint
from gpaw import GPAW, PoissonSolver
from gpaw.cluster import Cluster
from gpaw.test import equal


def test_generic_asym_box(in_tmp_dir):
    h = 0.2
    s = Cluster(Atoms('He'))
    s.minimal_box(3, h=h)

    c = GPAW(mode='fd', charge=1, txt='He_plus.txt',
             poissonsolver=PoissonSolver(use_charge_center=True),
             convergence={     # run fast
                 'energy': 0.001,
                 'eigenstates': 1e-4,
                 'density': 1e-3})
    s.calc = c
    e_small = s.get_potential_energy()
    eps_small = c.get_eigenvalues()[0]

    cell = s.get_cell()
    cell[0] *= 2
    s.set_cell(cell)
    e_big = s.get_potential_energy()
    eps_big = c.get_eigenvalues()[0]

    parprint('Energies and Eigenvalues:')
    parprint('     Small Box    Wide Box')
    parprint('E:   {0:9.3f}     {1:9.3f}'.format(e_small, e_big))
    parprint('eps: {0:9.3f}     {1:9.3f}'.format(eps_small, eps_big))
    equal(e_small, e_big, 2.5e-4)
    equal(eps_small, eps_big, 6e-4)



# http://www.answers.com/topic/hartree-units

from math import pi

from ase.units import _hbar, _eps0, _me, _e, _c, Bohr, Hartree, s

_a0 = Bohr * 1e-10
_autime = _hbar**3 * (4 * pi * _eps0)**2 / (_me * _e**4
                                            )  # 1 autime ~ 2.42e-17 s

# Conversion factors between ASE and GPAW units

attosec_to_autime = 1e-18 / _autime  # 1 as ~ 0.0413 autime
autime_to_attosec = _autime * 1e18  # 1 autime ~ 24.2 as

# ASE time unit is [Å sqrt(u/eV)]
# Hartree time unit is [hbar/Hartree]
asetime_to_autime = 1 / s / _autime  # 1 ASE time unit ~ 420.9 autime
autime_to_asetime = 1 / asetime_to_autime

eV_to_hartree = 1.0 / Hartree  # 1 eV ~ 0.0368 Eh
hartree_to_eV = Hartree  # 1 Eh ~ 27.2 eV

hartree_to_aufrequency = 1.0  # Eh ~ autime^(-1) since hbar=1
aufrequency_to_hartree = 1.0  # autime^(-1) ~ Eh since hbar=1

eV_to_aufrequency = 1.0 / Hartree  # 1 eV ~ 0.0368 autime^(-1)
aufrequency_to_eV = Hartree  # 1 autime^(-1) ~ 27.2 eV

eA_to_audipolemoment = 1 / Bohr  # 1 eÅ ~ 1.8897 aulength^(-1)
audipolemoment_to_eA = 1 / eA_to_audipolemoment

assert eV_to_aufrequency == eV_to_hartree * hartree_to_aufrequency
assert aufrequency_to_eV == aufrequency_to_hartree * hartree_to_eV

# Short-hand names
eV_to_au = eV_to_aufrequency
au_to_eV = aufrequency_to_eV
as_to_au = attosec_to_autime
au_to_as = autime_to_attosec
au_to_fs = au_to_as / 1e3
fs_to_au = as_to_au * 1e3
eA_to_au = eA_to_audipolemoment
au_to_eA = audipolemoment_to_eA

# Rotatory strength
# See https://doi.org/10.1016/0009-2614(95)01036-9
rot_au_to_SI = _e**2 * _hbar / _me * _a0  # 1 au = 1.6e-52 J C m / T
rot_au_to_cgs = rot_au_to_SI / (1e-6 / _c)  # 1 au = 4.7e-38 erg esu cm / gauss



from ase.spacegroup import crystal
from gpaw import GPAW
from gpaw import PW
from gpaw.test import equal


def test_symmetry_fractional_translations(in_tmp_dir):
    'sishovite'
    # no 136 - tetragonal

    a = 4.233944
    c = 2.693264
    p0 = (0, 0, 0)
    p1 = (0.306866, 0.306866, 0.0)

    atoms = crystal(['Si', 'O'], basis=[p0, p1],
                    spacegroup=136, cellpar=[a, a, c, 90, 90, 90])

    # with fractional translation
    calc = GPAW(mode=PW(),
                xc='LDA',
                kpts=(3, 3, 3),
                nbands=28,
                symmetry={'symmorphic': False},
                gpts=(18, 18, 12),
                eigensolver='rmm-diis',
                txt='with.txt')

    atoms.calc = calc
    energy_fractrans = atoms.get_potential_energy()

    assert len(calc.wfs.kd.ibzk_kc) == 6
    assert len(calc.wfs.kd.symmetry.op_scc) == 16

    # without fractional translations
    calc = GPAW(mode=PW(),
                xc='LDA',
                kpts=(3, 3, 3),
                nbands=28,
                gpts=(18, 18, 12),
                eigensolver='rmm-diis',
                txt='without.txt')

    atoms.calc = calc
    energy_no_fractrans = atoms.get_potential_energy()

    assert len(calc.wfs.kd.ibzk_kc) == 8
    assert len(calc.wfs.kd.symmetry.op_scc) == 8

    equal(energy_fractrans, energy_no_fractrans, 1e-5)



from abc import ABC, abstractmethod


class BaseSolver(ABC):
    """Abstract base class for solvers.

    Implementations of this class solves a set of linear equations A.x = b.

    Parameters
    ----------
    tolerance: float
        tolerance for the norm of the residual ||b - A.x||^2
    max_iterations: integer
        maximum number of iterations
    eps: float
        if abs(rho) or (omega) < eps, it's regarded as zero
        and the method breaks down

    Note
    ----
    Tolerance should not be smaller than attainable accuracy, which is
    order of kappa(A) * eps, where kappa(A) is the (spectral) condition
    number of the matrix. The maximum number of iterations should be
    significantly less than matrix size, approximately
    .5 sqrt(kappa) ln(2/tolerance). A small number is treated as zero
    if it's magnitude is smaller than argument eps.
    """
    def __init__(self,
                 tolerance=1e-8,
                 max_iterations=1000,
                 eps=1e-15):
        self.tol = tolerance
        self.max_iter = max_iterations
        if (eps <= tolerance):
            self.eps = eps
        else:
            raise ValueError(
                "Invalid tolerance (tol = %le < eps = %le)."
                % (tolerance, eps))

        self.iterations = -1

    def todict(self):
        return {'name': self.__class__.__name__,
                'tolerance': self.tol,
                'max_iterations': self.max_iter,
                'eps': self.eps}

    def initialize(self, gd, timer):
        """Initialize propagator using runtime objects.

        Parameters
        ----------
        gd: GridDescriptor
            grid descriptor for coarse (pseudowavefunction) grid
        timer: Timer
            timer
        """
        self.gd = gd
        self.timer = timer

    @abstractmethod
    def solve(self, A, x, b):
        """Solve a set of linear equations A.x = b.

        Parameters:
        A           matrix A
        x           initial guess x_0 (on entry) and the result (on exit)
        b           right-hand side (multi)vector

        """
        raise NotImplementedError()



"""Test of ScaLAPACK diagonalize and inverse cholesky.

The test generates a random symmetric matrix H0 and
positive definite matrix S0 on a 1-by-1 BLACS grid. They
are redistributed to a mprocs-by-nprocs BLACS grid,
diagonalized in parallel, and eigenvalues are compared
against LAPACK. Eigenvectors are not compared.
"""

import pytest
import numpy as np
from scipy.linalg import eigh, inv, cholesky

from gpaw.mpi import world, rank
from gpaw.blacs import BlacsGrid, Redistributor
from gpaw.utilities.tools import tri2full
from gpaw.utilities import compiled_with_sl
from gpaw.utilities.blas import rk
from gpaw.utilities.scalapack import \
    scalapack_general_diagonalize_dc, \
    scalapack_diagonalize_dc, \
    scalapack_inverse_cholesky, \
    scalapack_inverse, \
    scalapack_solve

from .test_pblas import \
    initialize_random, initialize_matrix, \
    calculate_error, mnprocs_i


pytestmark = pytest.mark.skipif(not compiled_with_sl(),
                                reason='not compiled with scalapack')


tol = 1.0e-8


@pytest.mark.parametrize('mprocs, nprocs', mnprocs_i)
@pytest.mark.parametrize('dtype', [float, complex])
def test_scalapack_diagonalize_inverse(dtype, mprocs, nprocs,
                                       N=72, seed=42):
    gen = np.random.RandomState(seed)
    grid = BlacsGrid(world, mprocs, nprocs)

    if dtype == complex:
        epsilon = 1.0j
    else:
        epsilon = 0.0

    # Create descriptors for matrices on master:
    glob = grid.new_descriptor(N, N, N, N)

    # print globA.asarray()
    # Populate matrices local to master:
    H0 = glob.zeros(dtype=dtype) + gen.rand(*glob.shape)
    S0 = glob.zeros(dtype=dtype) + gen.rand(*glob.shape)
    C0 = glob.empty(dtype=dtype)
    if rank == 0:
        # Complex case must have real numbers on the diagonal.
        # We make a simple complex Hermitian matrix below.
        H0 = H0 + epsilon * (0.1 * np.tri(N, N, k=-N // nprocs) +
                             0.3 * np.tri(N, N, k=-1))
        S0 = S0 + epsilon * (0.2 * np.tri(N, N, k=-N // nprocs) +
                             0.4 * np.tri(N, N, k=-1))
        # Make matrices symmetric
        rk(1.0, H0.copy(), 0.0, H0)
        rk(1.0, S0.copy(), 0.0, S0)
        # Overlap matrix must be semi-positive definite
        S0 = S0 + 50.0 * np.eye(N, N, 0)
        # Hamiltonian is usually diagonally dominant
        H0 = H0 + 75.0 * np.eye(N, N, 0)
        C0 = S0.copy()
        S0_inv = S0.copy()

    # Local result matrices
    W0 = np.empty((N), dtype=float)
    W0_g = np.empty((N), dtype=float)

    # Calculate eigenvalues / other serial results
    print(rank, C0.shape, C0.dtype)
    if rank == 0:
        W0 = eigh(H0, eigvals_only=True)
        W0_g = eigh(H0, S0, eigvals_only=True)
        C0 = inv(cholesky(C0, lower=True)).copy()  # result in lower triangle
        tri2full(S0_inv, 'L')
        S0_inv = inv(S0_inv)
        # tri2full(C0) # symmetrize

    print(rank, C0.shape, C0.dtype)
    assert glob.check(H0)
    assert glob.check(S0)
    assert glob.check(C0)

    # Create distributed destriptors with various block sizes:
    dist = grid.new_descriptor(N, N, 8, 8)

    # Distributed matrices:
    # We can use empty here, but end up with garbage on
    # on the other half of the triangle when we redistribute.
    # This is fine because ScaLAPACK does not care.

    H = dist.empty(dtype=dtype)
    S = dist.empty(dtype=dtype)
    Sinv = dist.empty(dtype=dtype)
    Z = dist.empty(dtype=dtype)
    C = dist.empty(dtype=dtype)
    Sinv = dist.empty(dtype=dtype)

    # Eigenvalues are non-BLACS matrices
    # W = np.empty((N), dtype=float)
    W_dc = np.empty((N), dtype=float)
    # W_mr3 = np.empty((N), dtype=float)
    # W_g = np.empty((N), dtype=float)
    W_g_dc = np.empty((N), dtype=float)
    # W_g_mr3 = np.empty((N), dtype=float)

    Glob2dist = Redistributor(world, glob, dist)
    Glob2dist.redistribute(H0, H, uplo='L')
    Glob2dist.redistribute(S0, S, uplo='L')
    Glob2dist.redistribute(S0, C, uplo='L')  # C0 was previously overwritten
    Glob2dist.redistribute(S0, Sinv, uplo='L')

    # we don't test the expert drivers anymore since there
    # might be a buffer overflow error
    # scalapack_diagonalize_ex(dist, H.copy(), Z, W, 'L')
    scalapack_diagonalize_dc(dist, H.copy(), Z, W_dc, 'L')
    # scalapack_diagonalize_mr3(dist, H.copy(), Z, W_mr3, 'L')
    # scalapack_general_diagonalize_ex(dist, H.copy(), S.copy(), Z, W_g, 'L')
    scalapack_general_diagonalize_dc(dist, H.copy(), S.copy(), Z, W_g_dc, 'L')

    scalapack_inverse_cholesky(dist, C, 'L')

    if dtype == complex:  # Only supported for complex for now
        scalapack_inverse(dist, Sinv, 'L')
    # Undo redistribute
    C_test = glob.empty(dtype=dtype)
    Sinv_test = glob.empty(dtype=dtype)
    Dist2glob = Redistributor(world, dist, glob)
    Dist2glob.redistribute(C, C_test)
    Dist2glob.redistribute(Sinv, Sinv_test)

    if rank == 0:
        diag_dc_err = abs(W_dc - W0).max()
        # diag_mr3_err = abs(W_mr3 - W0).max()
        # general_diag_ex_err = abs(W_g - W0_g).max()
        general_diag_dc_err = abs(W_g_dc - W0_g).max()
        # general_diag_mr3_err = abs(W_g_mr3 - W0_g).max()
        print(C_test[:2, :2])
        print(C0[:2, :2])
        inverse_chol_err = abs(C_test - C0).max()

        tri2full(Sinv_test, 'L')
        inverse_err = abs(Sinv_test - S0_inv).max()
        # print 'diagonalize ex err', diag_ex_err
        print('diagonalize dc err', diag_dc_err)
        # print 'diagonalize mr3 err', diag_mr3_err
        # print 'general diagonalize ex err', general_diag_ex_err
        print('general diagonalize dc err', general_diag_dc_err)
        # print 'general diagonalize mr3 err', general_diag_mr3_err
        print('inverse chol err', inverse_chol_err)
        if dtype == complex:
            print('inverse err', inverse_err)
    else:
        # diag_ex_err = 0.0
        diag_dc_err = 0.0
        # diag_mr3_err = 0.0
        # general_diag_ex_err = 0.0
        general_diag_dc_err = 0.0
        # general_diag_mr3_err = 0.0
        inverse_chol_err = 0.0
        inverse_err = 0.0

    # We don't like exceptions on only one cpu
    # diag_ex_err = world.sum(diag_ex_err)
    diag_dc_err = world.sum(diag_dc_err)
    # diag_mr3_err = world.sum(diag_mr3_err)
    # general_diag_ex_err = world.sum(general_diag_ex_err)
    general_diag_dc_err = world.sum(general_diag_dc_err)
    # general_diag_mr3_err = world.sum(general_diag_mr3_err)
    inverse_chol_err = world.sum(inverse_chol_err)
    inverse_err = world.sum(inverse_err)
    # assert diag_ex_err < tol
    assert diag_dc_err < tol
    # assert diag_mr3_err < tol
    # assert general_diag_ex_err < tol
    assert general_diag_dc_err < tol
    # assert general_diag_mr3_err < tol
    assert inverse_chol_err < tol
    if dtype == complex:
        assert inverse_err < tol


@pytest.mark.parametrize('mprocs, nprocs', mnprocs_i)
@pytest.mark.parametrize('dtype', [float, complex])
def test_scalapack_solve(dtype, mprocs, nprocs,
                         M=160, K=120, seed=42):
    """Test scalapack_solve()"""
    random = initialize_random(seed, dtype)
    grid = BlacsGrid(world, mprocs, nprocs)

    # Initialize matrices
    shapeA = (M, M)
    shapeB = (K, M)
    A0, A, descA = initialize_matrix(grid, *shapeA, 2, 2, random)
    B0, B, descB = initialize_matrix(grid, *shapeB, 3, 2, random)

    if grid.comm.rank == 0:
        # Calculate reference with numpy
        ref_B0 = np.linalg.solve(A0.T, B0.T).T
    else:
        ref_B0 = None

    # Calculate with scalapack
    scalapack_solve(descA, descB, A, B)

    # Check error
    err = calculate_error(ref_B0, B, descB)
    tol = {float: 8e-12, complex: 2e-13}[dtype]
    assert err < tol
